__all__ = ['CommanderNub', 'AuthCommanderNub', 'StdinNub', 'AuthStdinNub']

from tron import Misc, hub
from tron.Hub.Reply.ReplyTaster import ReplyTaster

from .CoreNub import CoreNub
from .NubAuth import NubAuth


class CommanderNub(CoreNub):
    """ Base class for ICC connections, where we accept commands from and s
    end replies to the remote end.  """

    def __init__(self, poller, **argv):
        """

        KWArgs:
           isUser    - if True, we should be listed as a logged-in user.
           forceUser - override any automatically derived username.
        """

        CoreNub.__init__(self, poller, **argv)

        # Note which Replies we want to accept. The default
        # is to accept only responses to our own commands.
        #
        self.taster = ReplyTaster()
        self.taster.setFilter((), (self.name, ), (self.name, ))

        self.isUser = argv.get('isUser', False)

        if 'forceUser' in argv:
            program, user = argv.get('forceUser').split('.')
            self.setNames(program, user)

    def __str__(self):
        return '%s(id=%s, name=%s, type=%s)' % (self.__class__.__name__, self.ID, self.name,
                                                self.nubType)

    def setNames(self, programName, username):
        """ Set our program and usernames. """

        Misc.log('CommandeNub.setNames',
                 'setting name for %s to %s.%s' % (self, programName, username))

        newName = hub.validateCommanderNames(self, programName, username)
        self.setName(newName)

    def setName(self, newName):
        """ Change our username(s). """

        self.name = newName
        self.encoder.setName(self.name)
        self.decoder.setName(self.name)

    def copeWithInput(self, s):
        """ Incorporate new input: buffer it, then extract and operate each complete new command.

        Args:
           s   - the new, but still unbuffered, input.

        Returns:
           Nothing.

        """

        if self.debug > 2:
            Misc.log('Nub.copeWithInput', 'CommanderNub %s read: %r' % (self.name, s))

        # Find and execute _every_ complete input.
        # The only time this function gets called is when new input comes in, so we
        # have no reliable mechanism for deferring input.
        #
        while True:
            cmd, leftover = self.decoder.decode(self.inputBuffer, s)
            s = None
            self.inputBuffer = leftover
            if cmd is None:
                break

            if self.log:
                try:
                    txt = cmd['RawText']
                except BaseException:
                    txt = 'UNKNOWN INPUT'
                self.log.log(txt, note='<')

            intercepted = False
            if hasattr(self, 'interceptCmd'):
                intercepted = self.interceptCmd(cmd)

            if not intercepted:
                hub.addCommand(cmd)

    def reply(self, r):

        # The authentication system want to be able to block all output until
        # the connection is established. Let an external agent "intercept" replies.
        #
        intercepted = False
        if hasattr(self, 'interceptReply'):
            intercepted = self.interceptReply(r)
        if intercepted:
            return

        # Most replies get sent to all interested commanders. But we allow
        # the possibility of only sending to the commander; in that case,
        # the commander gets all replies and keys, but other commanders only get told
        # about command completion.
        # We do this by triaging out replies here, then optionally telling the encoder
        # whether to include keys.
        #
        if r.bcast or r.cmd.cmdrID == self.ID:
            er = self.encoder.encode(r, self)
            self.queueForOutput(er)
            if self.log:
                self.log.log(er, note='>')
        else:
            Misc.log('CommanderNub.reply', 'not bcast; rID=%s selfID=%s' % (r.cmd.cmdrID, self.ID))
            if r.finishesCommand():
                er = self.encoder.encode(r, self, noKeys=True)
                self.queueForOutput(er)
                if self.log:
                    self.log.log(er, note='>')

    def tasteReply(self, r):
        if self.debug > 3:
            Misc.log('ActorNub.tasteReply', '%s tasting %s' % (self, r))

        if self.taster.taste(r):
            self.reply(r)

    def stopOutput(self):
        self.noOutput = True

    def startOutput(self):
        self.noOutput = False


class AuthCommanderNub(CommanderNub, NubAuth):
    """ A CommanderNub which enforces logins. """

    def __init__(self, poller, **argv):
        CommanderNub.__init__(self, poller, **argv)
        NubAuth.__init__(self, **argv)


class StdinNub(CommanderNub):

    def __init__(self, poller, in_f, out_f, **argv):
        CommanderNub.__init__(self, poller, **argv)

        self.mid = 1

        self.setInputFile(in_f)
        self.setOutputFile(out_f)


class AuthStdinNub(AuthCommanderNub):

    def __init__(self, poller, in_f, out_f, **argv):
        AuthCommanderNub.__init__(self, poller, **argv)

        self.mid = 1

        self.setInputFile(in_f)
        self.setOutputFile(out_f)



__all__ = ['keys']

from Vocab.InternalCmd import InternalCmd

from tron import Misc, g
from tron.Hub.KV.KVDict import kvAsASCII


class keys(InternalCmd):
    """ All the commands that the "keys" package provides.
        To wit:

        keys getFor=actor K1 [K2 [K3 ...]]

        Keywords returned:
           The requested keywords. But the src of the keywords is keys_actor
           instead of the actual actor.
    """

    def __init__(self, **argv):
        InternalCmd.__init__(self, 'keys', **argv)

        self.commands = {'getFor': self.getFor}

    def getFor(self, cmd, finish=True):
        """ Fetch keywords for a given actor

        Usage:
           getFor=actor K1 [K2 ... ]
        """

        matched, unmatched, leftovers = cmd.match([('getFor', str)])

        try:
            actor = matched['getFor']
        except KeyError:
            cmd.fail('keysTxt="no actor specified"')
            return

        try:
            g.KVs.sources[actor]
        except KeyError:
            cmd.fail('keysTxt="actor %s is not connected"' % (actor))
            return

        keys = list(leftovers.keys())
        Misc.log('keys.getFor',
                 'matched=%s; unmatched=%s; leftovers=%s' % (matched, unmatched, leftovers))

        matchedKeys, unmatchedKeys = g.KVs.getValues(actor, keys)
        if unmatchedKeys:
            failed = [Misc.qstr(x) for x in unmatchedKeys]
            cmd.warn('unmatchedKeys=%s' % (','.join(failed)), bcast=False)

        values = []
        for k, v in matchedKeys.items():
            values.append(kvAsASCII(k, v))

        if values:
            cmd.inform('; '.join(values),
                       noRegister=True,
                       src='keys_%s' % (actor),
                       bcast=False,
                       debug=9)
        if finish:
            cmd.finish(bcast=False)



import time

import tron.Hub.Command.Decoders as hubDecoders
import tron.Hub.Nub.Commanders as hubCommanders
import tron.Hub.Reply.Encoders as hubEncoders
from tron import g, hub
from tron.Hub.Nub.Listeners import SocketListener


name = 'client'
listenPort = 6093


def acceptStdin(in_f, out_f, addr=None):
    """ Create a command source with the given fds as input and output. """

    nubID = g.nubIDs.gimme()

    d = hubDecoders.ASCIICmdDecoder(needCID=True,
                                    needMID=True,
                                    EOL='\n',
                                    hackEOL=True,
                                    name=name,
                                    debug=1)
    e = hubEncoders.ASCIIReplyEncoder(EOL='\n', simple=True, debug=1, CIDfirst=True)
    c = hubCommanders.StdinNub(g.poller,
                               in_f,
                               out_f,
                               name='%s.v%d' % (name, nubID),
                               encoder=e,
                               decoder=d,
                               debug=1)

    c.taster.addToFilter(('*'), (), ('hub'))
    hub.addCommander(c)

    time.sleep(1)


def start(poller):
    stop()

    ll = SocketListener(poller, listenPort, name, acceptStdin)
    hub.addAcceptor(ll)

    time.sleep(1)


def stop():
    ll = hub.findAcceptor(name)
    if ll:
        hub.dropAcceptor(ll)
        del ll



__all__ = ['hubCommands']

import sys

import Vocab.InternalCmd as InternalCmd

from tron import Misc, g, hub
from tron.Hub.KV.KVDict import kvAsASCII


class hubCommands(InternalCmd.InternalCmd):
    """ All the commands that the "hub" package provides.

    The user executes these from the command window:

    hub startNubs tspec
    hub status
    etc.
    """

    def __init__(self, **argv):
        argv['safeCmds'] = r'^\s*(actors|commanders|actorInfo|version|status|ping)\s*$'
        argv['needsAuth'] = True
        InternalCmd.InternalCmd.__init__(self, 'hub', **argv)

        self.commands = {
            'actors': self.actors,
            'commanders': self.commanders,
            'restart!': self.reallyReallyRestart,
            'startNubs': self.startNubs,
            'stopNubs': self.stopNubs,
            'actorInfo': self.actorInfo,
            'commands': self.commandInfo,
            'setUsername': self.setUsername,
            'status': self.status,
            'loadWords': self.loadWords,
            'getKeys': self.getKeys,
            'listen': self.doListen,
            'version': self.version,
            'ping': self.status,
            'relog': self.relog,
        }

    def version(self, cmd, finish=True):
        """ Return the hub's version number. """

        hub.getSetHubVersion()

        vString = 'version=%s' % (g.KVs.getKV('hub', 'version', default='Unknown'))
        if finish:
            cmd.finish(vString)
        else:
            cmd.inform(vString)

    def doListen(self, cmd):
        """ Change what replies get sent to us. """

        matched, unmatched, leftovers = cmd.match([('listen', None), ('addActors', None),
                                                   ('delActors', None)])

        cmdr = cmd.cmdr()
        if not cmdr:
            cmd.fail('debug=%s' % (Misc.qstr('cmdr=%s; cmd=%s' % (cmdr, cmd))))
            return
        Misc.log('doListen', 'start: %s' % (cmdr.taster))
        Misc.log('doListen', 'leftovers: %s' % (leftovers))

        if 'addActors' in matched:
            actors = list(leftovers.keys())
            Misc.log('doListen', 'addActors: %s' % (actors))
            # cmd.inform('text="%s"' % (Misc.qstr("adding actors: %s" % (actors))))
            cmdr.taster.addToFilter(actors, [], actors)
            cmd.finish()
        elif 'delActors' in matched:
            actors = list(leftovers.keys())
            Misc.log('doListen', 'delActors: %s' % (actors))
            # cmd.inform('text="%s"' % (Misc.qstr("removing actors: %s" % (actors))))
            cmdr.taster.removeFromFilter(actors, [], actors)
            cmd.finish()
        else:
            cmd.fail('text="unknown listen command"')

        Misc.log('doListen', 'finish: %s' % (cmdr.taster))

    def actors(self, cmd, finish=True):
        """ Return a list of the currently connected actors. """

        g.actors.listSelf(cmd=cmd)
        if finish:
            cmd.finish('')

    def commanders(self, cmd, finish=True):
        """ Return a list of the currently connected commanders. """

        g.commanders.listSelf(cmd=cmd)
        if finish:
            cmd.finish('')

    def status(self, cmd, finish=True):
        Misc.cfg.flush()

        self.version(cmd, finish=False)
        self.actors(cmd, finish=False)
        self.commanders(cmd, finish=False)

        if finish:
            cmd.finish('')

    def setUsername(self, cmd):
        """ Change the username for the cmd's commander. """

        args = cmd.cmd.split()
        args = args[1:]

        if len(args) != 1:
            cmd.fail('cmdError="usage: setUsername newname"')
            return

        username = args[0]
        cmdr = cmd.cmdr()
        cmdr.setName(username)
        cmd.finish('')

    def stopNubs(self, cmd):
        """ stop a list of nubs. """

        nubs = list(cmd.argDict.keys())[1:]
        if len(nubs) == 0:
            cmd.fail('text="must specify one or more nubs to stop..."')
            return

        for nub in nubs:
            try:
                cmd.inform('text=%s' % (Misc.qstr('stopping nub %s' % (nub))))
                hub.stopNub(nub)
            except Exception as e:
                cmd.warn('text=%s' % (Misc.qstr('failed to stop nub %s: %s' % (nub, e))))

        cmd.finish('')

    def startNubs(self, cmd):
        """ (re-)start a list of nubs. """

        # Flush the configuration to force a reload later. This allows to change the
        # configuration or nubs during runtime without restarting tron.
        Misc.cfg.flush()

        nubs = list(cmd.argDict.keys())[1:]
        if len(nubs) == 0:
            cmd.fail('text="must specify one or more nubs to start..."')
            return

        for nub in nubs:
            try:
                cmd.inform('text=%s' % (Misc.qstr('(re-)starting nub %s' % (nub))))
                hub.startNub(nub)
            except Exception as e:
                cmd.warn('text=%s' % (Misc.qstr('failed to start nub %s: %s' % (nub, e))))

        cmd.finish('')

    def actorInfo(self, cmd):
        """ Get gory status about a list of actor nubs. """

        # Query all actors if none are specified.
        names = list(cmd.argDict.keys())[1:]
        if len(names) == 0:
            names = list(g.actors.keys())

        for n in names:
            try:
                nub = g.actors[n]
                nub.statusCmd(cmd, doFinish=False)
            except Exception as e:
                cmd.warn('text=%s' % (Misc.qstr('failed to query actor %s: %s' % (n, e))))

        cmd.finish('')

    def commandInfo(self, cmd):
        """ Get gory status about a list of actor nubs. """

        # Query all actors if none are specified.
        names = list(cmd.argDict.keys())[1:]
        if len(names) == 0:
            names = list(g.actors.keys())

        for n in names:
            try:
                nub = g.actors[n]
                nub.listCommandsCmd(cmd, doFinish=False)
            except Exception as e:
                cmd.warn('text=%s' % (Misc.qstr('failed to query actor %s: %s' % (n, e))))

        cmd.finish('')

    def loadWords(self, cmd, finish=True):
        """ (re-)load an internal vocabulary word. """

        words = list(cmd.argDict.keys())[1:]

        if len(words) == 0:
            words = None

        Misc.log('hubCmd', 'loadWords loading %s' % (words))
        try:
            hub.loadWords(words)
        except Exception as e:
            Misc.tback('hub.loadWords', e)
            cmd.fail('text=%s' % (Misc.qstr(e)))
            return

        if finish:
            cmd.finish()

    def getKeys(self, cmd):
        """ Return a bunch of keys for a given source.

        Cmd args:
            src  - a key source name.
            keys - 1 or more key names.
        """

        words = cmd.cmd.split()
        if len(words) < 3:
            cmd.fail('text="usage: getKeys srcName key1 [key2 ... keyN]"')
            return

        src = words[1]
        keys = words[2:]

        matched, unmatched = g.KVs.getValues(src, keys)
        Misc.log('hub.getKeys', 'matched=%s unmatched=%s' % (matched, unmatched))
        for k, v in matched.items():
            kvString = kvAsASCII(k, v)
            cmd.inform(kvString, src='hub.%s' % (src))
        if unmatched:
            cmd.warn('text=%s' % (Misc.qstr('unmatched %s keys: %s' %
                                            (src, ', '.join(unmatched)))))
        cmd.finish('')

    def reallyReallyRestart(self, cmd):
        """ Restart the entire MC. Which among other things kills us now. """

        cmd.warn('text=%s' %
                 (Misc.qstr('Restarting the hub now... bye, bye, and please call back soon!')))

        # Give the poller a chance to flush out the warning.
        g.poller.callMeIn(hub.restart, 1.0)

    def relog(self, cmd):
        """ Change where stderr goes to. """

        args = cmd.cmd.split()
        args = args[1:]

        if len(args) != 1:
            cmd.fail('cmdError="usage: relog filename"')
            return

        filename = args[0]
        import os

        f = open(filename, 'a', 1)
        os.dup2(f.fileno(), 1)
        os.dup2(f.fileno(), 2)
        sys.stdout = os.fdopen(1, 'w', 1)
        sys.stderr = os.fdopen(2, 'w', 1)
        f.close()

        cmd.finish('text="Jeebus, you done it now, whatever it was"')



__all__ = ['CommError']


class CommError(Exception):
    """ An exception that specifies that a low-level communication error occurred.
    These should only be thrown for serious communications errors. The top-level
    event loop will close/cleanup/destroy any running command. The error message
    will be returned on distxt.
    """

    def __init__(self, device, error, details=None):
        """ Create a CommError.

        Args:
           device  - name of the device that had an error.
           error   - one line of text, intended for users. Will be returned on distxt.
           details - optional text, intended for operators/programmers.
                     Will be returned on diserrtxt.
       """

        Exception.__init__(self)

        self.device = device
        self.error = error
        self.details = details
        if details:
            self.args = (device, error, details)
        else:
            self.args = (device, error)



from .ActorNub import *
from .Commanders import *
from .CoreNub import *
from .Listeners import *
from .NubAuth import *
from .RawActorNub import *
from .ShellNub import *
from .SocketActorNub import *
from .SocketCommanderNub import *
from .TCCShellNub import *



__all__ = ['CoreNub']

from tron import IO, Misc, hub


class CoreNub(IO.IOHandler):
    """ Base class for hub connections to the outside world.

    Because of the asymmetric nature of the protocol, each Nub must subclass from
    either CommanderNub or ActorNub, and not from CoreNub directly.

    - sendCommand(cmd)
    - sendReply(reply)
    - parseCommand(input)
    - parseReply(input)

    Each connection can/must be configured with specific:
       - input decoder  -- recognizes and extracts complete input chunks.
       - output encoder --
    """

    def __init__(self, poller, **argv):
        name = argv.get('name', None)
        self.ID = name
        self.name = name
        self.nubType = argv.get('type', 'unknown')
        self.needsAuth = argv.get('needsAuth', False)
        self.userInfo = None

        IO.IOHandler.__init__(self, poller, **argv)

        self.encoder = argv.get('encoder', None)
        self.decoder = argv.get('decoder', None)
        self.encoder.setNub(self.ID)
        self.decoder.setNub(self.ID)

        self.otherIP = argv.get('otherIP', None)
        self.otherFQDN = argv.get('otherFQDN', None)

        self.inputBuffer = ''
        self.outputBuffer = ''

        logDir = argv.get('logDir', None)
        if logDir:
            self.log = Misc.Logfile('nub.' + self.name, logDir)
        else:
            self.log = None

    def __str__(self):
        return 'CoreNub(id=%s, name=%s, type=%s)' % (self.ID, self.name, self.nubType)

    def setName(self, newName):
        """ Change our username(s). """

        self.name = newName
        self.encoder.setName(self.name)
        self.decoder.setName(self.name)

    def connected(self):
        pass

    def shutdown(self, **argv):
        """ Release all resources and shut down.

        Args:
           notifyHub(True)     - call the hub rather than clean up ourselves.
           why('')             - i

        If called from "below" (i.e. a socket has been shutdown), just
        calls the hub's dropNub method, which will shortly call us back.

        If called from the hub, close all IO resources.
        """

        notifyHub = argv.get('notifyHub', True)
        why = argv.get('why', '')

        Misc.log('Hub.shutdown', 'notify=%s why=%s' % (notifyHub, why))

        if notifyHub:
            hub.dropNub(self)
        else:
            self.ioshutdown(**argv)

    def flagFinishesCommand(self, f):
        """ Return True if a reply flag completes the command. """

        return f in ':fF'

    def statusCmd(self, cmd, doFinish=True):
        """ Send sundry status information keywords.
        """

        IO.IOHandler.statusCmd(self, cmd, self.name, doFinish=False)

        if doFinish:
            cmd.finish()



__all__ = ['parseArgs', 'match']

import re
from collections import OrderedDict

from tron import Misc

from .Exceptions import ParseException
from .keys import eatAString


# Match "  key = STUFF"
arg_re = re.compile(
    r"""
  ^\s*                          # Ignore leading space
  (?P<key>[a-z_][a-z0-9_-]*)    # Match keyword name
  \s*
  (?P<delimiter>=)
  \s*                           # Ignore spaces after keyname
  (?P<rest>.*)                  # Match eveything after the delimiter""",
    re.IGNORECASE | re.VERBOSE)
noarg_re = re.compile(
    r"""
  ^\s*                          # Ignore leading space
  (?P<key>\S+)                 # Match eveything up to the next WS
  \s*
  (?P<rest>.*)                  # Match eveything after the WS""", re.IGNORECASE | re.VERBOSE)


def eatAVee(s):
    """ Match a keyword value -- a possibly space-padded value ended by a
    whitespace, a comma, or a semicolon.

    Args:
       s - a string

    Returns:
       - the matched value. None if s is just whitespace.
       - any unmatched input, including the terminating character.
    """

    s = s.lstrip()
    if len(s) == 0:
        return '', ''

    # String parsing is trickier, let eatAString() handle that.
    if s[0] in "\"'":
        return eatAString(s)

    vEnd = len(s)
    for i in range(len(s)):
        if s[i] in ' \t\r\n\x0b\x0c':
            vEnd = i
            break

    return s[:vEnd], s[vEnd + 1:]


def parseArg(s):
    """ Try to parse a single KV.

    Return:
      { None, None, None } on end-of-input
      { K None rest-of-line } for a valueless keyword or
      { K V rest-of-line }
    """

    s = s.lstrip()
    if s == '':
        return None, None, None

    # Try to match for K=V. If we can't, gobble the next non-blank word.
    #
    match = arg_re.match(s)
    if match is None:
        match = noarg_re.match(s)
        if match is None:
            raise ParseException(leftoverText=s)
        d = match.groupdict()
        return d['key'], None, d['rest']

    d = match.groupdict()
    K = d['key']
    rest = d['rest']

    # Parse a value
    #
    try:
        val, rest = eatAVee(rest)
    except ParseException as e:
        e.prependText(rest)
        raise

    return K, val, rest


def parseArgs(s):
    """ Parse a string of command arguments into an OrderedDict .

    Returns:
      - an OrderedDict of keyword values.

    If a keyword has no value, the value is None
    Otherwise the value is a list of parsed values. Note that each value can be None.

    cmd a1 a2=1 a3= "2" a4=,
      ->

      args = {'a1' : None,
              'a2' : '1',
              'a3' : '2',
              'a4' : (None,None)
             }
    """

    KVs = OrderedDict()
    rest = s

    while True:
        try:
            key, values, rest = parseArg(rest)
        except ParseException as e:
            e.setKVs(KVs)
            raise

        if key is None:
            break

        KVs[key] = values

    # Misc.log('parseArgs', 'KVs: %s' % (KVs))
    return KVs


def match(argv, opts):
    """ Searches an OrderedDict for matches.

    Args:
      argv - an OrderedDict of options.
      opts - a list of duples to match against. The duple parts are the option name
             and a converter. If the converter is None, the option takes no argument.

    Returns:
      matches   - an OrderedDict of the matched options, with converted arguments.
      unmatched - a list of unmatched options from opts.
      leftovers - an OrderedDict of unmatched options from argv.

    Raises:
      Error     - Any parsing or conversion error.
    """

    # Convert the request duples to an OrderedDict
    want = OrderedDict()
    for o in opts:
        try:
            a, b = o
        except Exception:
            raise Exception('the argument to Command.matchDicts must be a list of duples')

        want[a] = b

    # Walk over the parsed options, and categorize them
    #
    matches = OrderedDict()
    leftovers = OrderedDict()

    for opt, arg in argv.items():
        # If we are looking for the option, match it and convert the argument.
        if opt in want:
            converter = want[opt]
            if converter is None:
                if arg is not None:
                    raise Exception('option %s takes no argument' % (Misc.qstr(opt, tquote="'")))
                matches[opt] = None
            else:
                try:
                    convArg = converter(arg)
                except Exception as e:
                    raise Exception("error with option '%s': %s" % (opt, e))

                matches[opt] = convArg

            # Remove the option from the search list.
            del want[opt]

        # If we are not looking for the option, return it as a leftover
        else:
            leftovers[opt] = arg

    return matches, list(want.keys()), leftovers



import os.path

from tron import g, hub
from tron.Hub.Command.Encoders.ASCIICmdEncoder import ASCIICmdEncoder
from tron.Hub.Nub.SocketActorNub import SocketActorNub
from tron.Hub.Reply.Decoders.ASCIIReplyDecoder import ASCIIReplyDecoder
from tron.Misc import cfg


name = 'alerts'


def start(poller):

    my_cfg = cfg.get(g.location, 'actors', doFlush=True)[name]

    stop()

    initCmds = ('ping', 'status')

    safeCmdsList = ['info', 'ping', 'version', 'status']
    safeCmds = r'^\s*({0})\s*$'.format('|'.join(safeCmdsList))

    d = ASCIIReplyDecoder(EOL='\r', stripChars='\n', CIDfirst=False, debug=1)
    e = ASCIICmdEncoder(EOL='\r', sendCommander=True, useCID=False, CIDfirst=False, debug=1)
    nub = SocketActorNub(poller, my_cfg['host'], my_cfg['port'],
                         name=name, encoder=e, decoder=d,
                         grabCID=True, # the actor spontaneously generates a line we can eat.
                         initCmds=initCmds, safeCmds=safeCmds,
                         needsAuth=True,
                         logDir=os.path.join(g.logDir, name),
                         debug=1)
    hub.addActor(nub)


def stop():
    n = hub.findActor(name)
    if n:
        hub.dropActor(n)
        del n



class cdict(object):
    """Dictionary, that has case-insensitive keys.

    Keys are retained in their original form
    when queried with .keys() or .items().

    Implementation: An internal dictionary maps lowercase
    keys to (key,value) pairs. All key lookups are done
    against the lowercase keys, but all methods that expose
    keys to the user retrieve the original keys.

    Taken from Sami Hangaslammi's ASPN submission:
        http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66315
    """

    def __init__(self, dict=None, dictType=dict):
        """Create an empty dictionary, or update from 'dict'."""
        self._dict = dictType()
        if dict:
            self.update(dict)

    def __contains__(self, key):
        """Case insensitive test where key is in dict"""
        k = key.lower()
        return k in self._dict

    def __delitem__(self, key):
        k = key.lower()
        del self._dict[k]

    def __getitem__(self, key):
        """Retrieve the value associated with 'key' (in any case)."""
        k = key.lower()
        return self._dict[k][1]

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        """Returns the number of (key, value) pairs."""
        return len(self._dict)

    def __repr__(self):
        """String representation of the dictionary."""
        items = ', '.join([('%r: %r' % (k, v)) for k, v in list(self.items())])
        return '{%s}' % items

    def __setitem__(self, key, value):
        """Associate 'value' with 'key'. If 'key' already exists, but
        in different case, it will be replaced."""
        k = key.lower()
        self._dict[k] = (key, value)

    def __str__(self):
        """String representation of the dictionary."""
        return repr(self)

    def clear(self):
        self._dict.clear()

    def fetch(self, key):
        """ Return both the cased key and the value. """

        k = key.lower()
        return self._dict[k]

    def get(self, key, default=None):
        """Retrieve value associated with 'key' or return default value
        if 'key' doesn't exist."""
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key, default):
        """If 'key' doesn't exists, associate it with the 'default' value.
        Return value associated with 'key'."""
        if key not in self:
            self[key] = default
        return self[key]

    def has_key(self, key):
        """Case insensitive test wether 'key' exists."""
        k = key.lower()
        return k in self._dict

    def items(self):
        """List of (key,value) pairs."""
        return list(self._dict.values())

    def iteritems(self):
        for key in self.keys():
            yield (key, self[key])

    def iterkeys(self):
        return iter(list(self.keys()))

    def itervalues(self):
        return iter(list(self.values()))

    def keys(self):
        """List of keys in their original case."""
        return [v[0] for v in list(self._dict.values())]

    def update(self, dict):
        """Copy (key,value) pairs from 'dict'."""
        for k, v in list(dict.items()):
            self[k] = v

    def values(self):
        """List of values."""
        return [v[1] for v in list(self._dict.values())]



__all__ = ['ASCIIReplyDecoder']

from tron import Misc
from tron.Parsing import parseASCIIReply

from .ReplyDecoder import ReplyDecoder


class ASCIIReplyDecoder(ReplyDecoder):

    def __init__(self, **argv):
        ReplyDecoder.__init__(self, **argv)

        self.EOL = argv.get('EOL', '\n')
        self.cidFirst = argv.get('CIDfirst', True)
        self.stripChars = argv.get('stripChars', '')

    def decode(self, buf, newData):
        """ Find and extract a single complete reply in the buf. Uses .EOL to
            recognize the end of a reply.

        Returns:
          - a Reply instance. None if .EOL no found in buf.
          - the content of buf with the first complete reply removed.

        Always consumes input up to the first .EOL, if .EOL is found.
        If .EOL is found, but the input can not be properly parsed, a modified reply is generated:

          - If no header information is found (i.e. no MID, CID, etc), the following is returned:
              w RawInput="full line"
          - If header information is found, but some part of the keywords are not parseable,
              x K1=V1; K2; RawKeys="rest of line"
          - If header information is found, and the last valueis an unterminated string, that
              value is silently terminated.
        """

        if newData:
            buf += newData

        if self.debug > 5:
            Misc.log('Stdin.extractReply', 'called with EOL=%r and buf=%r' % (self.EOL, buf))

        eol = buf.find(self.EOL)
        if self.debug > 4:
            Misc.log('Stdin.extractReply', 'eol at %d in buffer %r' % (eol, buf))

        # No complete reply found. make sure to return
        # the unmolested buffer.
        #
        if eol == -1:
            return None, buf

        replyString = buf[:eol]
        buf = buf[eol + len(self.EOL):]

        if self.debug > 2:
            Misc.log('Stdin.extractReply',
                     'hoping to parse (CIDfirst=%s) %r' % (self.cidFirst, replyString))

        for c in self.stripChars:
            replyString = replyString.replace(c, '')

        # Make sure to consume unparseable junk up to the next EOL.
        #
        try:
            r = parseASCIIReply(replyString, cidFirst=self.cidFirst)
        except SyntaxError as e:
            Misc.log('ASCIIReplyDecoder', 'Parsing error from %s: %r' % (self.name, e))
            return None, buf

        if self.debug > 3:
            Misc.log('Stdin.extractReply', 'extracted %r, returning %r' % (r, buf))

        return r, buf



__all__ = ['perms']

from tron import g

from .InternalCmd import InternalCmd


class perms(InternalCmd):
    """ All the commands that the "perms" package provides.
        To wit:

          perms status
          perms lock a1 a2 a3
          perms lock *
          perms unlock a1 a2 a3
          perms unlock *

          perms register PROGRAM
          perms drop PROGRAM

          perms CMD program=P

          CMD:
            set a1 a2 a3
            add a1 a2
            drop a1 a2

        Keywords returned:
          actors=a1,a2,a3
          authList=c,a1,a2,a3
          lockedActors=a1,a2
    """

    def __init__(self, **argv):
        argv['needsAuth'] = True
        argv['safeCmds'] = r'^\s*status\s*$'
        InternalCmd.__init__(self, 'perms', **argv)

        self.commands = {
            'status': self.status,
            'lock': self.lockCmd,
            'unlock': self.unlockCmd,
            'setLocked': self.setLockedCmd,
            'set': self.setCmd,
            'add': self.addCmd,
            'drop': self.dropCmd,
            'register': self.registerCmd,
            'unregister': self.unregisterCmd,
            'hackOn': self.hackOn,
            'hackOff': self.hackOff
        }

    def help(self, cmd, finish=True):
        pass

    def hackOn(self, cmd, finish=True):
        g.perms.hackOn = True
        cmd.warn('permsTxt="permissions are now disabled"')
        cmd.finish()

    def hackOff(self, cmd, finish=True):
        g.perms.hackOn = False
        cmd.warn('permsTxt="permissions are now in effect"')
        cmd.finish()

    def status(self, cmd, finish=True):
        """ Return the authentication keywords. """

        # Ignore any other arguments.
        #
        g.perms.status(cmd=cmd)

        if finish:
            cmd.finish()

    # auth.statusCmd.help = ('status', 'Return the authorization status keywords.', None)

    def setLockedCmd(self, cmd, finish=True):
        """ Lock a list of actors from non-APO commanders.

        Usage:
           setLocked a1 [a2 ...]
           setLocked *
        """

        args = list(cmd.argDict.keys())[1:]
        if len(args) == 1 and args[0] == '*':
            actors = None
        else:
            actors = args

        g.perms.setLockedActors(actors, cmd=cmd)

        if finish:
            cmd.finish()

    def lockCmd(self, cmd, finish=True):
        """ Lock a list of actors from non-APO commanders.

        Usage:
           lock a1 [a2 ...]
           lock *
        """

        args = list(cmd.argDict.keys())[1:]
        if len(args) == 1 and args[0] == '*':
            actors = []
        else:
            actors = args

        g.perms.lockActors(actors, cmd=cmd)

        if finish:
            cmd.finish()

    def unlockCmd(self, cmd, finish=True):
        """ unlock a list of actors from non-APO commanders.

        Usage:
           unlock a1 [a2 ...]
           unlock *
        """

        args = list(cmd.argDict.keys())[1:]

        if len(args) == 1 and args[0] == '*':
            actors = []
        else:
            actors = args

        g.perms.unlockActors(actors, cmd=cmd)

        if finish:
            cmd.finish()

    def registerCmd(self, cmd, finish=True):
        """ Register a list of programs to control.

        Usage:
           register p1 [p2 ...]
           register *
        """

        args = list(cmd.argDict.keys())[1:]
        if len(args) == 0:
            cmd.fail('text="perms register requires one or more program names"')
            return

        if len(args) == 1 and args[0] == '*':
            programs = []
        else:
            programs = args

        g.perms.addPrograms(programs, cmd=cmd)

        if finish:
            cmd.finish()

    def unregisterCmd(self, cmd, finish=True):
        """ Unregister a list of programs to control.

        Usage:
           unregister p1 [p2 ...]
           unregister *
        """

        args = list(cmd.argDict.keys())[1:]
        if len(args) == 0:
            cmd.fail('text="perms unregister requires one or more program names"')
            return

        if len(args) == 1 and args[0] == '*':
            programs = []
        else:
            programs = args

        g.perms.dropPrograms(programs, cmd=cmd)

        if finish:
            cmd.finish()

    def setCmd(self, cmd, finish=True):
        """ Define the actors that a given program can command.

        Usage:
           set prog=p [a1 ...]
        """

        matched, unmatched, leftovers = cmd.match([('set', None), ('program', str)])

        try:
            program = matched['program']
        except KeyError:
            cmd.fail('authTxt="no program specified"')
            return

        g.perms.setActorsForProgram(program, leftovers, cmd=cmd)

        if finish:
            cmd.finish()

    def addCmd(self, cmd, finish=True):
        """ Add to the actors that a given program can command.

        Usage:
           add program=p [a1 ...]
        """

        matched, unmatched, leftovers = cmd.match([('add', None), ('program', str)])

        try:
            program = matched['program']
        except KeyError:
            cmd.fail('authTxt="no program specified"')
            return

        g.perms.addActorsToProgram(program, leftovers, cmd=cmd)

        if finish:
            cmd.finish()

    def dropCmd(self, cmd, finish=True):
        """ Block some actors that a given program can command.

        Usage:
           drop program=p [a1 ...]
        """

        matched, unmatched, leftovers = cmd.match([('drop', None), ('program', str)])

        try:
            program = matched['program']
        except KeyError:
            cmd.fail('authTxt="no program specified"')
            return

        g.perms.dropActorsFromProgram(program, leftovers, cmd=cmd)

        if finish:
            cmd.finish()



__all__ = ['ASCIICmdEncoder']

from tron import Misc

from .CommandEncoder import CommandEncoder


class ASCIICmdEncoder(CommandEncoder):
    """ Encode commands into the simple ASCII protocol.

    Options:
        EOL:       specify the EOL string. Default is '\n'
        useCID:    whether we can meaningfully specify the CID. Uses
                   0 for the CID if False. Default is True.
        useTarget: whether the name of the Actor should be included. Default is False.
    """

    def __init__(self, **argv):
        CommandEncoder.__init__(self, **argv)
        self.EOL = argv.get('EOL', '\n')
        self.useCID = argv.get('useCID', True)
        self.CIDfirst = argv.get('CIDfirst', True)
        self.useTarget = argv.get('useTarget', False)
        self.sendCmdr = argv.get('sendCommander', False)
        self.sendCmdrCID = argv.get('sendCommanderCID', False)

    def encode(self, cmd):
        if self.useCID:
            if self.CIDfirst:
                ids = '%s %s ' % (cmd.actorCid, cmd.actorMid)
            else:
                ids = '%s %s ' % (cmd.actorMid, cmd.actorCid)
        else:
            ids = '%s ' % (cmd.actorMid, )

        if self.sendCmdrCID:
            cmdrInfo = '%s ' % (cmd.cmdrCid)
        elif self.sendCmdr:
            cmdrInfo = '%s ' % (cmd.cmdrName)
        else:
            cmdrInfo = ''

        if self.useTarget:
            e = '%s%s %s%s%s' % (cmdrInfo, cmd.actorName, ids, cmd.cmd, self.EOL)
        else:
            e = '%s%s%s%s' % (cmdrInfo, ids, cmd.cmd, self.EOL)

        if self.debug > 5:
            Misc.log('ASCIIEncoder', 'encoded: %s' % (e))

        return e



__all__ = ['PyReplyDecoder']

import pickle

from tron import Misc

from .ReplyDecoder import ReplyDecoder


class PyReplyDecoder(ReplyDecoder):
    """ Encode Replys as single-line pickled python objects.
    """

    def __init__(self, **argv):
        ReplyDecoder.__init__(self, **argv)

        # How do we terminate encoded lines?
        #
        self.EOL = argv.get('EOL', '\f')

    def decode(self, buf, newData):
        """ Find and extract a single complete command in the inputBuffer.
        """

        if newData:
            buf += newData

        if self.debug > 3:
            Misc.log('PyReply.decoder', 'called with EOL=%r and buf=%r' % (self.EOL, buf))

        eol = buf.find(self.EOL)
        if self.debug > 2:
            Misc.log('PyReply.decoder', 'eol at %d in buffer %r' % (eol, buf))

        # No complete reply found. make sure to return
        # the unmolested buffer.
        #
        if eol == -1:
            return None, buf

        replyString = buf[:eol]
        buf = buf[eol + len(self.EOL):]

        # Make sure to consume unparseable junk up to the next EOL.
        #
        try:
            r = pickle.loads(replyString)
        except SyntaxError:
            Misc.log('PyReply.decoder', 'Failed to unpickle %r' % (replyString))
            return None, buf

        if self.debug > 5:
            Misc.log('PyReply.decoder', 'extracted %r, returning %r' % (r, buf))

        return r, buf



"""
Start the hub server, including all Nubs for connections.
"""

import sys

from tron import Misc, g, hub


def startAllConnections(names):
    """ Create all default connections, as defined by the proper configuration file. """

    for n in names:
        try:
            hub.startNub(n)
        except Exception as e:
            msg = 'FAILED to start nub %s: %s\n' % (n, e)
            sys.stderr.write(msg)
            try:
                g.hubcmd.warn('text=%s' % (Misc.qstr(msg)))
            except BaseException:
                sys.stderr.write('hubcmd.warn failed\n')


# NOTE: jkp: I don't like it, but I'll work with the "everything's global" setup
# for now to store the location.
# TODO: hub should be a class we init() anyway!
g.location = Misc.location.determine_location()
print(f'Detected location: {g.location}.')

hub.init()
startAllConnections(Misc.cfg.get(g.location, 'nubs', doFlush=True))

hub.run()



__all__ = ['getDayDirName', 'getQuarterName']

import time


localTZ = 6 * 3600


def getDayDirName(t=None):
    """ Return a per-day directory name.

    Returns:
         - a string of the form "UT041219"

    The rule is a bit odd:
       - the night starts at local noon
       -
    """

    if t is None:
        t = time.time()
    localNow = t - localTZ

    localNowPlus12H = localNow + (12 * 3600)
    dateString = time.strftime('UT%y%m%d', time.gmtime(localNowPlus12H))

    return dateString


def getQuarterName(t=None):
    """ Return the current quarter name.

    Returns:
      - a string of the form 'Q3'
    """

    if t is None:
        t = time.time()
    localNow = t - localTZ

    # Uck. We do _not_ want the last night in the quarter to be assigned to the next quarter.
    # So use _today_'s date to determine the quarter.
    localNowMinus12H = localNow - (12 * 3600)
    month = time.gmtime(localNowMinus12H)[1]

    return 'Q%d' % ((month + 2) / 3)


def _test():
    now = time.time()
    for h in range(24):
        testNow = now + h * 3600
        print(getDayDirName(t=testNow))


if __name__ == '__main__':
    _test()



#!/usr/bin/env python

__all__ = ['FileHandler']

import sys

from tron import Misc

from . import IOHandler


class FileHandler(IOHandler):
    """ Class for IO connections that are connected to files/pipes. """

    def __init__(self, poller, in_f=None, out_f=None, inputCallback=None):
        IOHandler.__init__(self, poller, in_f, out_f)

        self.inputCallback = inputCallback

    def readInput(self):
        # Misc.log("Cat.readInput", "reading...")
        try:
            readIn = self.in_f.read(self.tryToRead)
        except IOError:
            Misc.log('Cat.readInput', 'exception %r' % (sys.exc_info, ))
            self.poller.removeInput(self)
            if len(self.outBuffer) > 0:
                self.poller.removeOutput(self)
            return
        except BaseException:
            self.poller.removeInput(self)
            if len(self.outBuffer) > 0:
                self.poller.removeOutput(self)
            raise

        if len(readIn) == 0:
            Misc.log('FileHandler.readInput',
                     'READ 0 BYTES from %s .... CEASING TO READ!' % (self, ))
            self.poller.removeInput(self)

        Misc.log('Cat.readInput', 'read len=%d %r' % (len(readIn), readIn[:30]))
        self.inBuffer += readIn

        if self.inputCallback:
            self.inputCallback(self)

    def mayOutput(self):
        Misc.log('Cat.mayOutput', 'writing len=%d %r' % (len(self.outBuffer), self.outBuffer[:30]))

        try:
            sent = self.out_f.write(self.outBuffer[:self.tryToWrite])
        except IOError as e:
            Misc.log('Cat.mayOutput', 'exception %r' % (e, ))
            self.poller.removeOutput(self)
            try:
                self.poller.removeInput(self)
            except BaseException:
                pass
            return
        except BaseException:
            raise

        self.outBuffer = self.outBuffer[sent:]
        if len(self.outBuffer) == 0:
            self.poller.removeOutput(self)



__all__ = ['SocketCommanderNub']

import socket

from .Commanders import CommanderNub


class SocketCommanderNub(CommanderNub):

    def __init__(self, poller, host, port, **argv):
        """
        """

        CommanderNub.__init__(self, poller, **argv)
        self.host = host
        self.port = port

        f = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        f.connect((host, port))
        f.setblocking(0)

        self.setInputFile(f)
        self.setOutputFile(f)

        self.connected()



__all__ = ['BinaryReplyDecoder']

import os
import re
import struct
import tempfile
from collections import OrderedDict

import pixel16  # This does not exist but it seems this module is not imported.

from tron import Misc, g
from tron.Parsing import ParseException, parseKVs

from .ReplyDecoder import ReplyDecoder


class BinaryReplyDecoder(ReplyDecoder):
    msg_re = re.compile(
        r"""
    \s*                          # Skip leading whitespace
    (?P<flag>[iw:f>!])           # The flag. Should allow more characters,
                                 # and check them elsewhere.
    (?P<rest>.*)""", re.VERBOSE | re.IGNORECASE)

    def __init__(self, scratchDir, **argv):
        ReplyDecoder.__init__(self, **argv)

        # Where do we save scratch data.
        #
        self.scratchDir = scratchDir
        self.imageBuf = None

        # Whether to byte-swap or sign-flip the data.
        self.doByteSwapFirst = argv.get('byteSwapFirst', False)
        self.doByteSwapLast = argv.get('byteSwapLast', False)
        self.doSignFlip = argv.get('signFlip', False)
        self.BZERO = argv.get('BZERO', 0.0)

    def saveImageData(self, image, xpix, ypix, bitpix):
        """ Save image data to a scratch FITS file, register and return the file name. """

        f, fname = tempfile.mkstemp('.fits', '%s-' % (self.nubID), self.scratchDir)

        # Create a minimal header.
        hdr = []
        fmt = '%-08s=%21s / %-47s'
        hdr.append(fmt % ('SIMPLE', 'T', ''))
        hdr.append(fmt % ('BITPIX', repr(bitpix), 'Number of bits/data pixel'))
        hdr.append(fmt % ('NAXIS', '2', 'An image'))
        hdr.append(fmt % ('NAXIS1', repr(xpix), 'The number of columns'))
        hdr.append(fmt % ('NAXIS2', repr(ypix), 'The number of rows'))

        if self.BZERO != 0.0:
            hdr.append(fmt % ('BSCALE', 1.0, ''))
            hdr.append(fmt % ('BZERO', repr(self.BZERO), ''))

        hdr.append('%-80s' % ('END'))

        hdr_s = ''.join(hdr)
        remain = len(hdr_s) % 2880
        os.write(f, hdr_s + ' ' * (2880 - remain))

        # Possibly fiddle the data bits.
        if self.doByteSwapFirst:
            Misc.log('Binary.saveImage', 'byteswap first')
            pixel16.byteswap(image)
        if self.doSignFlip:
            Misc.log('Binary.saveImage', 'signflip')
            pixel16.uflip(image)
        if self.doByteSwapLast:
            Misc.log('Binary.saveImage', 'byteswap last')
            pixel16.byteswap(image)

        written = os.write(f, image)
        assert (written == len(image)
                ), 'SHORT WRITE ON IMAGE DATA: written=%d len(image)=%d' % (written, len(image))

        # Pad data to FITS 2880-byte block.
        remain = len(image) % 2880

        if remain > 0:
            Misc.log('Binary.saveImage',
                     'padding %d-byte data with %d null bytes' % (len(image), 2880 - remain))
            os.write(f, '\000' * (2880 - remain))

        os.close(f)

        # Register the filename. We probably eventually want to register the
        # data, but this is safer.
        #
        g.KVs.setKV('images', '%sFile' % (self.nubID), fname, None)

        return fname

    def decode(self, buf, newData):

        if newData:
            buf += newData

        # The binary protocol encapsulates each message in a 10-byte header and a 2-byte trailer:
        #
        # 1(1 byte)
        #        I have no idea what this is about.
        # isfile(1 byte)
        #        Nor this, really. In the hub, if the value is > 1, the data is saved to a file.
        # length(4 bytes)
        #        Length of the complete packet, including the header and trailer.
        # mid(2 bytes) cid(2 bytes)
        #        Note the range: 0..64k
        #        We don't use the cid.
        # if (isfile <= 1):
        #     message
        # else:
        #     xpix(2 bytes)
        #     ypix(2 bytes)
        #     bitpix(2 bytes) (or 1?)
        #     data
        #
        # checksum(1 byte)
        #        1-byte XOR of the message body.
        # 4(1 byte)
        #        No idea. Just send 4.

        # Quick check for minimal length.
        #
        if len(buf) < 12:
            return None, buf

        # Examine first part, especially the length
        #
        dummy, is_file, length, cid, mid = \
            struct.unpack('>BBihh', buf[:10])
        if dummy != 1:
            # Complain, but don't fail.
            Misc.log('Hub.decap', 'dummy=%d is_file=%d length=%d mid=%d cid=%d' %
                     (dummy, is_file, length, mid, cid))
        is_file = is_file > 1

        if self.debug >= 5:
            Misc.log('Binary.decap', 'is_file=%s length=%d (%d) cid=%d mid=%d' %
                     (is_file, length, len(buf), cid, mid))

        fullLength = length + 10 + 2

        if len(buf) < fullLength:
            return None, buf

        # We don't have a complete command yet. Keep waiting.
        #
        if is_file:
            headerLength = 16
        else:
            headerLength = 10

        if is_file:
            xpix, ypix, bitpix = struct.unpack('>hhh', buf[10:16])
        msg = buf[headerLength:fullLength - 2]

        # Trailer parts.
        csum, trailer = struct.unpack('>BB', buf[fullLength - 2:fullLength])

        # Calculate & check checksum of message body.
        #   Because images come through here, we need to C this. -- Misc
        #
        my_csum = 0
        if not is_file:
            for i in range(10, fullLength - 10 + 1):
                my_csum ^= ord(buf[i])
            if my_csum != csum:
                Misc.log('Hub.decap', 'csum(%d) != calculated csum(%d)' % (csum, my_csum))

        # Magic trailer value. I don't know what this means, but ctrl-d can be Unix EOF.
        #
        if trailer != 4:
            Misc.error('Hub.decap', 'trailer is not 4 (%d)' % (trailer))
            Misc.log('Hub.decap', "mid=%d cid=%d len=%d msg='%s'"
                     % (mid, cid, length, msg))

        buf = buf[fullLength:]

        if self.debug >= 7:
            Misc.log('Binary.decap', 'csum=%d match=%s trailer=%d left=%d (%r) msg=(%r)' %
                     (csum, csum == my_csum, trailer, len(buf), buf, msg))
        elif self.debug >= 5:
            Misc.log('Binary.decap', 'csum=%d match=%s trailer=%d left=%d (%r)' %
                     (csum, csum == my_csum, trailer, len(buf), buf))

        d = {'mid': mid, 'cid': cid}

        if is_file:
            d['flag'] = 'i'
            d['rest'] = ''
            KVs = OrderedDict()
            KVs['xpix'] = xpix
            KVs['ypix'] = ypix
            KVs['bitpix'] = bitpix
            KVs['scratchFile'] = self.saveImageData(msg, xpix, ypix, bitpix)

            d['KVs'] = KVs
        else:
            match = self.msg_re.match(msg)
            if match is None:
                d['flag'] = 'w'
                KVs = OrderedDict()
                KVs['UNPARSEDTEXT'] = Misc.qstr(msg)
                d['KVs'] = KVs

                return d, buf

            msg_d = match.groupdict()

            d['flag'] = msg_d['flag']
            d['rest'] = msg_d['rest']

            try:
                KVs = parseKVs(msg_d['rest'])
            except ParseException as e:
                KVs = e.KVs
                KVs['UNPARSEDTEXT'] = Misc.qstr(e.leftoverText)
            except Exception as e:
                Misc.log('parseASCIIReply', 'unexpected Exception: %s' % (e))
                KVs = OrderedDict()
                KVs['RawLine'] = Misc.qstr(msg_d['rest'])

            d['KVs'] = KVs

        return d, buf



__all__ = ['qstr']


def qstr(o, equotes=None, tquote='"'):
    r""" Put a string representation of an object into quotes and escape it minimally.

    Return the string wrapped in tquotes.
    Escape all the characters in equotes, as well as backslashes. If equotes are
    are not defined, use tquote.

    Basically,
       o  -> "str(o)"
       \  -> \\
       eq  -> \eq
    """

    s = str(o)

    # Always quote backslashes _first_.
    #
    if equotes is None:
        if tquote is None:
            return s
        equotes = '\\' + tquote
    else:
        equotes = '\\' + equotes

    # Should compare with a clever RE scheme:
    #   matches = match_all(equotes)
    #   '\\'.join(match pieces)
    #
    for equote in equotes:
        equote_repl = '\\' + equote
        idx = 0
        while True:
            dq = s.find(equote, idx)
            #        sys.stderr.write("dq=%d idx=%d len=%d\n s=%r" % (dq, idx, len(s), s))
            if dq == -1:
                break

            s = ''.join((s[:dq], equote_repl, s[dq + 1:]))
            idx = dq + 2

    if tquote:
        return ''.join((tquote, s, tquote))
    else:
        return s


if __name__ == '__main__':
    tests = ('', '"', '""', "''", "'", "\'", '\"', '\\', '\"', 'abcdef', 'abc"\\')

    for t in tests:
        qt = qstr(t)
        try:
            e = eval(qt)
        except Exception:
            print('===== NE: %s' % (t))
            print('        : %s' % (qt))
            continue

        if t == e:
            print('===== OK: %s' % (qt))
        else:
            print('===== NG: %s:' % (t))
            print('        : %s:' % (qt))

    print()
    print()



__all__ = ['TCCShellNub']

from tron import Misc
from tron.Hub import Reply

from .ShellNub import ShellNub


class TCCShellNub(ShellNub):
    """ An ShellNub with the plumbing required to recognize and record the TCC YourUserName as
        our .cid.
    """

    def findUserNum(self, kvl):
        """ Find YourUserNum key in list of KVs. Return the CID or None. """

        for k, v in kvl.items():
            if k == 'YourUserNum':
                cid = int(v[0])
                return cid
        return None

    def copeWithInput(self, s):
        """ Override the default copeWithInput to set our .cid from the YourUserNum key. """

        if self.debug > 5:
            Misc.log('TCCShell.copeWithInput',
                     'Nub %s read: %r, with buf=%r' % (self.name, s, self.inputBuffer))

        while True:
            # Connections to the TCC's tccuser captive account return lines
            # terminated by CRLF, but with the LF coming at the start of the "next
            # line". Odd, and to be investigated. In the meanwhile, strip leading LFs
            #
            if len(self.inputBuffer) > 0 and self.inputBuffer[0] == '\n':
                self.inputBuffer = self.inputBuffer[1:]

            reply, leftover = self.decoder.decode(self.inputBuffer, s)
            s = None
            if self.debug > 5:
                Misc.log('TCCShell.copeWithInput',
                         'decoded: %s, yielding buf=%r' % (reply, leftover))

            self.inputBuffer = leftover
            if not reply:
                break

            if self.log:
                try:
                    txt = reply['RawText']
                except BaseException:
                    txt = 'UNKNOWN INPUT'
                self.log.log(txt, note='<')

            # Here's the special TCC bit: search for YourUserNum,
            if self.cid is None:
                newCID = self.findUserNum(reply['KVs'])
                if newCID is not None:
                    self.cid = newCID
                    Misc.log('TCCShell.copeWithInput', 'setting CID=%s' % (self.cid))
                    self.connected()

            cmd = self.getCmdForReply(reply)
            r = Reply.Reply(cmd, reply['flag'], reply['KVs'])
            cmd.reply(r)



__all__ = ['ShellNub']

import os
import resource
import signal
import sys

from tron import Misc

from .ActorNub import ActorNub


class ShellNub(ActorNub):

    def __init__(self, poller, cmd, **argv):
        ActorNub.__init__(self, poller, **argv)

        self.sig = signal.SIGHUP
        self.shell(cmd)

    def ioshutdown(self, **argv):
        """ Reap our dead child. """

        ActorNub.ioshutdown(self, **argv)

        try:
            os.kill(self.pid, self.sig)
        except Exception as e:
            Misc.log('Shell.shutdown',
                     'os.kill(pid=%s, sig=%s) failed with %s' %
                     (self.pid, self.sig, e))

        pid, status = os.waitpid(self.pid, 0)
        Misc.log('Shell.shutdown', 'waitpid returned pid=%s and status=%s' % (pid, status))

    def shell(self, cmd):

        Misc.log('Shell.shell', '%s launching %r' % (self.name, cmd))

        self.cmd = cmd
        p1_i, p1_o = os.pipe()
        p2_i, p2_o = os.pipe()

        pid = os.fork()
        if pid == 0:
            # Child

            os.close(p2_i)
            os.close(p1_o)

            os.dup2(p1_i, 0)
            os.dup2(p2_o, 1)
            os.dup2(sys.stderr.fileno(), 2)

            # Close the rest of the file descriptors
            #
            (fd_max_soft, fd_max_hard) = resource.getrlimit(resource.RLIMIT_NOFILE)
            for fd in range(3, fd_max_soft):
                try:
                    os.close(fd)
                except BaseException:
                    pass

            os.execv(cmd[0], cmd)
            assert 'child' == 'alive'

        else:
            # Parent

            os.close(p1_i)
            os.close(p2_o)

            self.pid = pid
            if self.name is None:
                self.ID = 'shell-%ld' % (pid)
                self.name = self.ID

            self.setInputFile(os.fdopen(p2_i, 'r'))
            self.setOutputFile(os.fdopen(p1_o, 'w'))

            Misc.log('Shell.shell', "launched '%s' %r as pid %d" % (cmd[0], cmd[1:], pid))



import socket
import time

from tron import g, hub
from tron.Hub.Command.Decoders.ASCIICmdDecoder import ASCIICmdDecoder
from tron.Hub.Nub.Commanders import AuthStdinNub
from tron.Hub.Nub.Listeners import SocketListener
from tron.Hub.Reply.Encoders.ASCIIReplyEncoder import ASCIIReplyEncoder


name = 'TUI'
listenPort = 9877


def acceptTUI(in_f, out_f, addr=None):
    """ Create a command source with the given fds as input and output. """

    # Fetch a unique ID
    #
    nubID = g.nubIDs.gimme()
    fullname = '%s_%d' % (name, nubID)

    # all = ('tcc','mcp',
    #       'hub','msg')
    all = ('*', )

    otherIP, otherPort = in_f.getpeername()
    try:
        otherFQDN = socket.getfqdn(otherIP)
    except BaseException:
        otherFQDN = 'unknown'

    # os.system("/usr/bin/sudo /usr/local/bin/www-access add %s" % (otherIP))

    d = ASCIICmdDecoder(needCID=False, EOL='\r\n', debug=1)
    e = ASCIIReplyEncoder(EOL='\r', simple=True, debug=1, CIDfirst=True)
    c = AuthStdinNub(g.poller,
                     in_f,
                     out_f,
                     name=fullname,
                     encoder=e,
                     decoder=d,
                     debug=1,
                     type='TUI',
                     needsAuth=True,
                     isUser=True,
                     otherIP=otherIP,
                     otherFQDN=otherFQDN)
    c.taster.addToFilter(all, (), all)
    hub.addCommander(c)


def start(poller):
    stop()

    lt = SocketListener(poller, listenPort, name, acceptTUI)
    hub.addAcceptor(lt)


def stop():
    a = hub.findAcceptor(name)
    if a:
        hub.dropAcceptor(a)
        del a
        time.sleep(0.5)  # OK, why did I put this in here?



__all__ = ['RawCmdDecoder']

from tron import Misc
from tron.Hub.Command import Command

from .CommandDecoder import CommandDecoder


class RawCmdDecoder(CommandDecoder):
    """ A Command decoder for accepting commands which have no target, MID, or CID. We
    know our target, and assign an incrementing MID. In other words, we transform:

       cmdTxt -> tgt mid cmdTxt

    """

    def __init__(self, target, **argv):
        """ Create ourself.

        Args:
           target     - the name of an Actor.
        Optargs:
           EOL        - the EOL string (default='\n')
           cmdWrapper - if set, call this command at the target actor, and pass the incoming
                        cmd as an argument.
        """
        CommandDecoder.__init__(self, **argv)

        self.target = target
        self.EOL = argv.get('EOL', '\n')
        self.CID = argv.get('CID', '0')
        self.stripChars = argv.get('stripChars', '')
        self.cmdWrapper = argv.get('cmdWrapper', None)
        self.mid = 1

        if self.debug > 1:
            Misc.log('RawCmdDecoder.init',
                     'target=%s cmdWrapper=%s' % (self.target, self.cmdWrapper))

    def decode(self, buf, newData):
        """ Find and extract a single complete command from the given buffer.

        Returns:
           - a Command instance, or None if no complete command was found.
           - the unconsumed part of the buffer.

           If a command-sized piece is found, but cannot be parsed,
           return None, leftovers.

        """

        if newData:
            buf += newData

        eol = buf.find(self.EOL)

        if self.debug > 3:
            Misc.log('RawCmdDecoder.extractCmd', 'EOL at %d in buffer %r' % (eol, buf))

        # No complete command found. Return the original buffer so that the caller
        # can easily determine that no input was consumed.
        #
        if eol == -1:
            return None, buf

        # We have a complete command. Strip it off from the rest of the input buffer.
        #
        cmdString = buf[:eol]
        buf = buf[eol + len(self.EOL):]

        for c in self.stripChars:
            cmdString = cmdString.replace(c, '')
        if self.debug > 3:
            Misc.log('RawCmdDecoder.extractCmd', 'cmdString=%r' % (cmdString))

        if self.cmdWrapper:
            cmdString = '%s raw=%s' % (self.cmdWrapper, cmdString)

        # Assign a new MID
        #
        mid = self.mid
        self.mid += 1

        return Command(self.nubID, self.CID, mid, self.target, cmdString), buf



__all__ = ['tcmd']


class tcmd(object):

    def __init__(self, name='tcmd'):
        self.name = name

    def inform(self, *args):
        print('%s.inform: %s' % (self.name, ''.join(*args)))

    def warn(self, *args):
        print('%s.warn: %s' % (self.name, ''.join(*args)))

    def fail(self, *args):
        print('%s.fail: %s' % (self.name, ''.join(*args)))



#!/usr/bin/env python

__all__ = ['ID']

import threading


class ID(object):
    """ Provide a thread-safe counter.
    """

    def __init__(self, start=1):
        self.id = start
        self.lock = threading.Lock()

    def gimme(self):
        """ Return the next ID, and increment the internal counter. """

        self.lock.acquire()
        id = self.id
        self.id += 1
        self.lock.release()

        return id



#!/usr/bin/env python

import os

from tron import Misc


location = Misc.location.determine_location()


def getActors(actorName=None, hostName=None):
    # Bootstrap the whole configuration system
    configPath = os.environ.get('CONFIG_DIR', os.path.join(os.environ['TRON_DIR'], 'config'))
    Misc.cfg.init(path=configPath, verbose=False)

    actors = Misc.cfg.get(location, 'actors')
    if actorName:
        actors = dict([(key, val) for key, val in list(actors.items()) if key == actorName])
    if hostName:
        actors = dict([(key, val) for key, val in list(actors.items())
                       if (hostName in (val['host'], val['host'].split('.', 1)[0]))])
    return actors


def printActors(actors, verbose=False):
    if verbose:
        actorList = [
            '%s,%s,%s' % (actors[name]['actorName'], actors[name]['host'], actors[name]['port'])
            for name in list(actors.keys())
        ]
    else:
        actorList = [actors[name]['actorName'] for name in list(actors.keys())]

    if actorList:
        print('\n'.join(actorList))


def main():
    printActors(getActors())


if __name__ == '__main__':
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option('-H',
                      '--host',
                      dest='hostName',
                      default=None,
                      help='match actors on given host',
                      metavar='HOST')
    parser.add_option('-n',
                      '--name',
                      dest='actorName',
                      default=None,
                      help='match given actor',
                      metavar='ACTOR')
    parser.add_option('-v',
                      '--verbose',
                      action='store_true',
                      dest='verbose',
                      default=False,
                      help='print all actor fields')

    (options, args) = parser.parse_args()

    actors = getActors(actorName=options.actorName, hostName=options.hostName)
    printActors(actors, verbose=options.verbose)



__all__ = ['RawCmdEncoder']

from tron import Misc

from .CommandEncoder import CommandEncoder


class RawCmdEncoder(CommandEncoder):
    """ Encode commands into a raw ASCII protocol: one without any MID, CID, etc.

    Options:
        EOL:       specify the EOL string. Default is '\n'
    """

    def __init__(self, **argv):
        CommandEncoder.__init__(self, **argv)

        self.EOL = argv.get('EOL', '\n')

    def encode(self, cmd):

        e = '%s%s' % (cmd.cmd, self.EOL)
        if self.debug > 5:
            Misc.log('RawEncoder', 'encoded: %s' % (e))

        return e



__all__ = ['LLock']

from threading import Lock

from tron import Misc


class LLock(object):
    """ Debugging Lock. Can print when the acquire & release calls are made. """
    seq = 1

    def __init__(self, debug=0, name=None):
        self.debug = debug
        self.lock = Lock()
        if name is None:
            name = 'lock-%04d' % (self.seq)
            self.seq += 1
        self.name = name

        if self.debug > 0:
            Misc.log('LLock.create', 'name=%s' % (self.name))

    def acquire(self, block=True, src='up'):
        if self.debug > 0:
            Misc.log('LLock.acquiring', 'name=%s, block=%s, src=%s' %
                     (self.name, block, src))
        self.lock.acquire(block)
        if self.debug > 0:
            Misc.log('LLock.acquired', 'name=%s, block=%s, src=%s' %
                     (self.name, block, src))

    def release(self, src='up'):
        if self.debug > 0:
            Misc.log('LLock.release', 'name=%s, src=%s' %
                     (self.name, src))
        self.lock.release()



__all__ = ['setID', 'setLogdir', 'enableLoggingFor', 'disableLoggingFor',
           'isoTS', 'log', 'error']

import os
from math import modf
from time import gmtime, strftime, time


systems = {}

DISABLED = '0'
ENABLED = '.'
ERROR = 'E'
FATAL = 'F'
UNDEFINED = '?'

logfileDir = os.environ["TRON_LOG_DIR"]
logfileName = None
logID = 'log'
logfile = None
rolloverOffset = -0.3 * (24 * 3600)  # Fermi MJD offset.
rolloverChunk = 24 * 3600
rolloverTime = 0


def setID(newID):
    global logID

    logID = newID


def setLogdir(dirname):
    global logfileDir

    logfileDir = dirname


def enableLoggingFor(system):
    systems[system] = ENABLED


def disableLoggingFor(system):
    systems[system] = DISABLED


def setLoggingFor(system, level):
    if level:
        systems[system] = ENABLED
    else:
        systems[system] = DISABLED


def isoTS(t=None, format='%Y-%m-%d %H:%M:%S', zone='Z'):
    """ Return a proper ISO timestamp for t, or now if t==None. """

    if t is None:
        t = time()

    if zone is None:
        zone = ''

    return strftime(format, gmtime(t)) \
        + '.%03d%s' % (1000 * modf(t)[0], zone)


def rollover(t):
    global logfile
    global logfileName
    global rolloverTime

    if t > rolloverTime:
        logfile = None

    if logfile is None:
        # Set next rollover time.
        rolloverTime = t - t % rolloverChunk + rolloverChunk + rolloverOffset

        # this can happen at startup
        if rolloverTime < t:
            rolloverTime += rolloverChunk
        logfileName = '%s.log' % (strftime('%Y-%m-%dT%H:%M:%S', gmtime(t)))
        os.makedirs(os.path.join(logfileDir, logID), exist_ok=True)
        logfile = open(os.path.join(logfileDir, logID, logfileName), 'w', 1)
        currentName = os.path.join(logfileDir, logID, 'current.log')
        try:
            os.unlink(currentName)
            os.symlink(logfileName, currentName)
        except BaseException:
            pass
        log('log', 'next rollover is at %d (%s)' % (rolloverTime, isoTS(rolloverTime)))


def log(system, detail, state=None):
    now = time()
    nowStamp = isoTS(now)
    rollover(now)

    # if not hasattr(globals(), 'logfile'):
    #    logfile = sys.stderr

    # If the logging state has not explicitely been enabled or disabled,
    # print the notice, but mark the system name with a '?'
    #
    if state is None:
        state = systems.get(system, UNDEFINED)

    if state == UNDEFINED:
        state = systems.get('default', state)

    if state != DISABLED:
        logfile.write('%s %s %s %s %s\n' % (nowStamp, logID, state, system, detail))
        logfile.flush()


def error(*args):
    log(*args, **{'state': ERROR})



__all__ = ['ParseException']


class ParseException(Exception):
    """ A mini Exception used to pass useful information up from the bowels of the parser.

    In particular:
        leftoverText is whatever text has not been parsed.
        KVs is whatever keys have already been parsed.

    """

    def __init__(self, leftoverText, KVs=None):
        Exception.__init__(self)

        self.leftoverText = leftoverText
        self.KVs = KVs

    def __str__(self):
        return 'ParseException(leftover=%s, KVs=%s)' % (self.leftoverText, self.KVs)

    def prependText(self, t):
        """ Insert t at the beginning of .leftoverText. """

        self.leftoverText = t + self.leftoverText

    def setKVs(self, KVs):
        """ Set .KVs. """

        self.KVs = KVs



#!/usr/bin/env python

__all__ = ['SocketHandler']

import sys

from tron import Misc

from . import IOHandler


class SocketHandler(IOHandler):
    """ Class for IO connections that are connected to a socket. """

    def __init__(self, poller, in_f=None, out_f=None):
        IOHandler.__init__(self, poller, in_f, out_f)

        self.debug = 1

    def readInput(self):
        # Misc.log("Cat.readInput", "reading...")
        try:
            readIn = self.in_f.recv(self.tryToRead)
        except IOError:
            Misc.log('Cat.readInput', 'exception %r' % (sys.exc_info, ))
            self.poller.removeInput(self)
            if len(self.outBuffer) > 0:
                self.poller.removeOutput(self)
            return
        except BaseException:
            self.poller.removeInput(self)
            if len(self.outBuffer) > 0:
                self.poller.removeOutput(self)
            raise

        if self.debug:
            Misc.log('Cat.readInput', 'read len=%d %r' % (len(readIn), readIn[:30]))
        self.inBuffer += readIn

    def mayOutput(self):
        if self.debug:
            Misc.log('Cat.mayOutput',
                     'writing len=%d %r' % (len(self.outBuffer), self.outBuffer[:30]))

        try:
            sent = self.out_f.send(self.outBuffer[:self.tryToWrite])
        except IOError as e:
            Misc.log('Cat.mayOutput', 'exception %r' % (e, ))
            self.poller.removeOutput(self)
            try:
                self.poller.removeInput(self)
            except BaseException:
                pass
            return
        except BaseException:
            raise

        self.outBuffer = self.outBuffer[sent:]
        if len(self.outBuffer) == 0:
            self.poller.removeOutput(self)



__all__ = ['Error']


class Error(Exception):

    def __init__(self, oneliner, **argv):
        """ Create a generic error with arbitray attributes.

        Args:
           error   - one line of text, intended for users.
           argv    - a dictionary of attributes.
        """

        Exception.__init__(self)

        self.oneliner = oneliner

        self.argv = argv
        self.args = oneliner

    def __getattr__(self, name):
        try:
            return self.argv[name]
        except BaseException:
            raise AttributeError('%s instance has no attribute %s' %
                                 (self.__class__.__name__, name))



"""Utilities for dealing with our location (APO vs. LCO)."""

import os
import socket
import warnings


def determine_location(location=None):
    """Return a location based on the domain name."""

    location = location or os.environ.get('OBSERVATORY', None)

    if location is None:
        fqdn = socket.getfqdn().split('.')
    else:
        location = location.upper()
        assert location in ['APO', 'LCO', 'LOCAL'], 'invalid location'
        return location

    if 'apo' in fqdn:
        return 'APO'
    elif 'lco' in fqdn:
        return 'LCO'
    elif 'ACTORCORE_LOCAL' in os.environ and os.environ['ACTORCORE_LOCAL'] == '1':
        return 'LOCAL'
    else:
        warnings.warn('Using local setup for tron.', UserWarning)
        return 'LOCAL'



__all__ = ['ASCIIReplyEncoder']

from tron import Misc
from tron.Hub.KV.KVDict import kvAsASCII
from tron.Parsing.keys import parseKVs

from .ReplyEncoder import ReplyEncoder


class ASCIIReplyEncoder(ReplyEncoder):
    """ Encode Replys in a basic ASCII protocol. """

    def __init__(self, **argv):
        ReplyEncoder.__init__(self, **argv)

        # How do we terminate encoded lines?
        #
        self.EOL = argv.get('EOL', '\n')

        # Do we encode the minimal info, or include full tracking info?
        #
        self.simple = argv.get('simple', True)
        if self.simple:
            self.encode = self.encodeSimple
        else:
            self.encode = self.encodeFull

        # And should the key source be included?
        #
        self.noSrc = argv.get('noSrc', False)

        self.CIDfirst = argv.get('CIDfirst', False)

    def encodeSimple(self, r, nub, noKeys=False):
        """ Encode a reply for a given nub.

        The simple encoding returns the minimum that a selfish ICC needs to know about --
        whether the reply is to one of its commands, and if so which one.
        """

        cmd = r.cmd
        cid = cmd.cmdrCid
        mid = cmd.cmdrMid

        # This is scary. For clients which do not identify their command sources, we want to return
        # their CID (0) unmolested but we need to name the client to other listeners.
        #
        if cid == '0' and self.nubID != cmd.cmdrName:
            cid = cmd.cmdrName

        if self.CIDfirst:
            id_s = '%s %s' % (cid, mid)
        else:
            id_s = '%s %s' % (mid, cid)

        if self.debug > 5:
            Misc.log('ASCIIReplyEncoder.encode',
                     'CIDfirst=%s, cid=%s, mid=%s, id_s=%s' % (self.CIDfirst, cid, mid, id_s))

        if noKeys:
            keys = ''
        else:
            keys = self.encodeKeys(r.src, r.KVs)

        if self.noSrc:
            return '%s %s %s%s' % (id_s,
                                   r.flag, keys, self.EOL)
        else:
            return '%s %s %s %s%s' % (id_s,
                                      r.src, r.flag, keys, self.EOL)

    def encodeFull(self, r, nub, noKeys=False):
        """ Encode a reply for a given nub.

        Encode all the information required to track the source of the command and the reply.
        """

        cmd = r.cmd

        if noKeys:
            keys = ''
        else:
            keys = self.encodeKeys(r.src, r.KVs)

        return '%s %s %s %s %s %s %s %s %s%s' % (cmd.cmdrName, cmd.cmdrMid, cmd.cmdrCid,
                                                 cmd.actorName, cmd.actorMid, cmd.actorCid, r.src,
                                                 r.flag, keys, self.EOL)

    def encodeKeys(self, src, KVs):
        """ Return a string encoding of KVs stored in an OrderedDict.

        Args:
           src   - ignored
           KVs   - an OrderedDict of values. See Parsing/parsing.py for important details.
        Notes:

        """

        if self.debug > 5:
            Misc.log('ASCIIReplyEnc.encode', 'encoding %r' % (KVs, ))
        if KVs is None:
            return ''

        keylist = []
        for k, v in KVs.items():
            if self.debug > 5:
                Misc.log('ASCIIReplyEnc.encode', 'encoding %r=%r' % (k, v))

            keylist.append(kvAsASCII(k, v, escape=self.EOL))

        return '; '.join(keylist)


if __name__ == '__main__':
    import sys

    tests = ('Received="26437.910 00010000   14.6028678 -10.32"   0"',
             "distxt=\"Unrecognized command: '\"'\"", '', 'e="', "e='",
             'abc = "oh no", "please no", 1', 'd=99.9', 'e=99,9', "k=a,'b'", 'z29=   abc;  def',
             'k9="sddsas')
    sys.stderr.write('\n\n')
    for t in tests:
        sys.stderr.write('====== :%s:\n' % (t))
        d = parseKVs(t)
        print('t=:%s: d=%s' % (t, d))



__all__ = ['Reply']

import time
from collections import OrderedDict

from tron import Misc, Parsing


"""
   Reply is a slight misnomer --

   - Reply to an existing command. Need to specify:
       - flag, cmd, [src], KVs
   - Generate non-command KVs. Need to specify:
       - flag, actorCid, actorMid, src, KVs
"""


class Reply(Misc.Object):

    def __init__(self, cmd, flag, KVs, bcast=True, **argv):
        """ Create a parsed Reply.

        Args:
           cmd  - the Command which we are a Reply to.
           flag - the completion state flag.
           KVs  - parsed or unparsed keys. We accept OrderedDicts,
                  lists & tuples, and strings. The latter are parsed
                  into OrderedDicts.
        """

        Misc.Object.__init__(self, **argv)

        self.ctime = time.time()
        self.cmd = cmd
        self.flag = flag
        self.bcast = bcast

        if isinstance(KVs, OrderedDict):
            self.KVs = KVs
        else:
            self.KVs = self.parseKVs(KVs)

        self.src = argv.get('src', cmd.actorName)

    def finishesCommand(self):
        """ Return true if the given flag finishes a command. """

        return self.flag in ':fF'

    def __str__(self):
        return 'Reply(cmd=%s flag=%s KVs=%s)' % (self.cmd, self.flag, self.KVs)

    def parseKVs(self, kvl):
        """ Convert some form of keys to an OrderedDict.

        We are trying to be ridiculously flexible here. Take:

         - a string, which we parse as it came from an ICC.
         - a list, which we parse either as a list of key=value strings or of (key, value) duples.
        """

        if isinstance(kvl, str):
            return Parsing.parseKVs(kvl)

        od = OrderedDict()
        if kvl is not None:
            for i in kvl:
                if isinstance(i, str):
                    k, v, junk = Parsing.parseKV(i)
                    od[k] = v
                elif type(i) in (list, tuple) and len(i) == 2:
                    k, v, junk = Parsing.parseKV('%s=%s' % i)
                else:
                    Misc.log('Reply', 'kvl item is not a string: %r' % (i))
                    raise Exception('kvl == %r' % (i))

        return od



__all__ = ['ReplyDecoder']

from tron import Misc


class ReplyDecoder(Misc.Object):
    """ Base class for decoders for incoming replies.

    """

    def __init__(self, **argv):
        Misc.Object.__init__(self, **argv)

        self.name = argv.get('name', 'unnamed')
        self.nubID = None

    def setNub(self, n):
        self.nubID = n

    def setName(self, s):
        self.name = s

    def decode(self, s0, s1):
        raise RuntimeError('.decode() must be defined in a ReplyDecoder subclass.')



__all__ = ['InternalCmd']

import re

from tron import Misc


class InternalCmd(object):

    def __init__(self, name, isActor=True, **argv):
        """ The common parts of internal actors.

        Args:
           name         - our public name

        Optional Args:
           isActor      - whether we should be listed as an Actor

        """

        self.name = name
        self.ID = name
        self.isActor = isActor
        self.needsAuth = argv.get('needsAuth', False)
        # if self.needsAuth == True:
        #     self.needsAuth = self.name

        self.locked = False

        self.debug = argv.get('debug', 0)
        self.commands = {}

        safeCmds = argv.get('safeCmds', None)
        if safeCmds:
            self.safeCmds = re.compile(safeCmds)
        else:
            self.safeCmds = None

        # Some stats
        self.totalCommands = 0

    def sendCommand(self, cmd):
        """
        """

        Misc.log('%s.cmd' % (self.name), 'running cmd=%s' % (Misc.qstr(cmd.cmd)))
        self.totalCommands += 1

        cmd.parseArgs()
        words = list(cmd.argDict.keys())
        if len(words) == 0:
            cmd.finish('')
            return

        cmdWord = words[0]
        cmdHandler = self.commands.get(cmdWord, None)
        if cmdHandler is None:
            cmd.fail('%sTxt=%s' %
                     (self.name, Misc.qstr('No command named %s' % (cmdWord))))
            return

        cmd.reportQueued()
        try:
            cmdHandler(cmd)
        except Exception as e:
            Misc.tback('Vocab.sendCommand', e)
            cmd.fail('%sTxt=%s' % (self.name, Misc.qstr(e, tquote='"')))
            return

    def statusCmd(self, cmd, doFinish=True):
        """ """

        cmd.inform('vocabStats=%s,%d' % (Misc.qstr(self.name), self.totalCommands))

        if doFinish:
            cmd.finish()

    def listCommandsCmd(self, cmd, doFinish=True):
        """ """

        if doFinish:
            cmd.finish()

    def shutdown(self, notifyHub=None):
        pass



__all__ = ['RawReplyDecoder']

from tron import Misc
from tron.Parsing import parseRawReply

from .ReplyDecoder import ReplyDecoder


class RawReplyDecoder(ReplyDecoder):

    def __init__(self, **argv):
        ReplyDecoder.__init__(self, **argv)

        self.EOL = argv.get('EOL', '\n')
        self.stripChars = argv.get('stripChars', '')

    def decode(self, buf, newData):
        """ Find and extract a single complete reply in the buf. Uses .EOL to
            recognize the end of a reply.

        Returns:
          - a Reply instance. None if .EOL no found in buf.
          - the content of buf with the first complete reply removed.

        Always consumes input up to the first .EOL, if .EOL is found.
        """

        if newData:
            buf += newData

        if self.debug > 5:
            Misc.log('Stdin.extractReply', 'called with EOL=%r and buf=%r' % (self.EOL, buf))

        eol = buf.find(self.EOL)
        if self.debug > 4:
            Misc.log('Stdin.extractReply', 'eol at %d in buffer %r' % (eol, buf))

        # No complete reply found. make sure to return
        # the unmolested buffer.
        #
        if eol == -1:
            return None, buf

        replyString = buf[:eol]
        buf = buf[eol + len(self.EOL):]

        if self.debug > 2:
            Misc.log('Stdin.extractReply', 'hoping to parse %r' % (replyString))

        for c in self.stripChars:
            replyString = replyString.replace(c, '')

        r = parseRawReply(replyString)

        if self.debug > 3:
            Misc.log('RawReplyDecoder.extractReply', 'extracted %r, returning %r' % (r, buf))

        return r, buf



#!/usr/bin/env python

"""
   The core of the hub.
   The hub basically:
      - maintains a keyword-value ("KV") dictionary,
      - accepts commands from Commanders and passes them onto Actors,
      - accepts and generates KV responses, and passes them on to interested parties.

   We handle the fact that we can get (or generate) responses to commands that that we
   did not send by creating Commands for all responses.

   c = Command()
      cmdrCid, cmdrMid
      tgt, cmd
      [xid]

      actorCid, actorMid

   r = Reply(cmd=cmd

"""

import importlib
import json
import os
import re
import signal
import sys
import time
from collections import OrderedDict

import tron.Auth
import tron.Hub.Command.Command
import tron.Hub.KV.KVDict
import tron.IO
from tron import Misc, __version__, g
from tron.Misc.cdict import cdict


def init(programsFile=None):
    g.home = sys.path[0]
    if g.home is None or g.home == '':
        g.home = os.getcwd()

    g.rootDir = os.getcwd()

    if 'TRON_PACKAGE_DIR' in os.environ:
        tron_package_dir = os.environ['TRON_PACKAGE_DIR']
    else:
        tron_package_dir = os.path.dirname(__file__)

    # Bootstrap the whole configuration system
    configPath = os.environ.get('CONFIG_DIR', os.path.join(tron_package_dir, 'config'))

    Misc.cfg.init(path=configPath)
    os.environ['CONFIG_DIR'] = configPath

    g.logDir = os.path.expandvars(Misc.cfg.get('hub', 'logDir'))
    Misc.setLogdir(g.logDir)
    Misc.setID('hub')
    Misc.log('hub.init', 'logger started...')

    #   - a globally unique ID generator for Commands.
    g.xids = Misc.ID()
    g.nubIDs = Misc.ID()
    g.hubMIDs = Misc.ID()

    # The Hub is basically:
    #   All of these are in the global namespace "g".
    #
    #   - A dictionary of KVs
    g.KVs = tron.Hub.KV.KVDict.KVDict(debug=5)

    g.commanders = cdict()
    g.actors = cdict()

    g.hubcmd = None
    g.hubcmd = tron.Hub.Command.Command(
        '.hub',
        '0',
        0,
        'hub',
        None,
        actorCid=0,
        actorMid=0,
        neverEnd=True)

    #   - An authorization manager
    permsCmd = tron.Hub.Command.Command(
        '.perms',
        '0',
        0,
        'perms',
        None,
        actorCid=0,
        actorMid=0,
        neverEnd=True)
    g.perms = tron.Auth.Auth(permsCmd, debug=1)

    # Loads the programs with their custom permissions
    Misc.log('hub.init', 'loading programs ...')
    loadPrograms()

    #   - A dictionary of Commander Nubs, indexed by unique ID.
    g.commanders = CmdrDict('Commanders')
    # g.listeners = g.commanders

    #   - A dictionary of Actor Nubs, indexed by name
    g.actors = NubDict('Actors')
    g.vocabulary = cdict()

    #   - dictionary of PollAcceptors, waiting for for new connections.
    g.acceptors = cdict()

    #   - dictionary of active commands, indexed by XID.
    g.pendingCommands = {}

    #   - A PollHandler
    g.poller = tron.IO.PollHandler(debug=1)

    Misc.log('hub.init', 'loading internal vocabulary...')
    loadWords(None)

    Misc.log('hub.init', 'loading keys...')
    loadKeys()

    # atexit.register(shutdown)
    signal.signal(signal.SIGHUP, handleSIGHUP)
    signal.signal(signal.SIGTERM, handleSIGTERM)


def handleSIGHUP(signal, frame):
    restart()


def handleSIGTERM(signal, frame):
    shutdown()


def getSetHubVersion():
    """ Put the uncached svn version info into the hub.version keyword. """

    version = Misc.qstr(__version__)
    g.KVs.setKV('hub', 'version', version, None)


def loadKeys():
    rootDir = Misc.cfg.get(g.location, 'httpRoot')
    host = Misc.cfg.get(g.location, 'httpHost')

    g.KVs.setKV('hub', 'httpRoot', [host, rootDir], None)

    getSetHubVersion()


def loadWords(words=None):
    if words is None:
        words = Misc.cfg.get('hub', 'vocabulary')

    for w in words:
        _loadWords([w])


def _loadWords(wordlist):
    """ (Re-)load a list of Vocabulary words, overwriting any existing info.
    """

    # First, (re-)load the entire Vocabulary module. Let that fail to the top
    # level.
    #
    Misc.log('hub.loadVocab', 'trying to (re-)load Vocab module')

    for w in wordlist:
        # Now try to load the module itself.
        modName = 'tron.Vocab.' + w
        if w == 'hub':
            modName = 'tron.Vocab.hubCommands'
        try:
            Misc.log('hub.loadVocab', 'trying to (re-)load vocabulary word %s' % (w,))
            mod = importlib.import_module(modName)
        except ImportError as e:
            raise Exception('Import of %s failed: %s' % (modName, e))

        Misc.log('hub.loadWords', 'loading vocabulary word %s from %s...' % (w, modName))

        cmdSet = getattr(mod, 'hubCommands' if w == 'hub' else w)()
        try:
            dropActor(cmdSet)
        except BaseException:
            pass
        g.vocabulary[w] = cmdSet

        addActor(cmdSet)

        Misc.log('hub.loadWords', 'vocabulary: %s' % (g.vocabulary))


def shutdown():
    Misc.log('hub.shutdown', 'shutting down......................................')
    try:
        _shutdown()
    except BaseException:
        pass
    sys.exit(0)


def restart():
    Misc.log('hub.restart', 'restarting......................................')
    try:
        _shutdown()
    except BaseException:
        pass

    Misc.log('hub.restart', 'for real......................................')
    time.sleep(1)
    os.execlp('tron', 'tron', 'restart')


def _shutdown():
    sys.stderr.write('final cleanup; deleting hub pieces...\n')

    sys.stderr.write('       deleting acceptors...\n')
    for aname, acceptor in list(g.acceptors.items()):
        try:
            acceptor.shutdown(notifyHub=False)
        except BaseException:
            pass

    sys.stderr.write('       deleting commanders...\n')
    for cname, cmdr in list(g.commanders.items()):
        try:
            cmdr.shutdown(notifyHub=False)
        except BaseException:
            pass

    sys.stderr.write('       deleting actors...\n')
    for aname, actor in list(g.actors.items()):
        try:
            actor.shutdown(notifyHub=False)
        except BaseException:
            pass


def run():
    """ Listens for and handles I/O on all devices.
    """
    while True:
        try:
            Misc.log('hub.run', 'actors id=%r' % (id(g.actors)))
            g.poller.run()
        except (SystemExit, KeyboardInterrupt):
            Misc.log('Hub.run', 'Normal exit."')
            raise

        except Exception as e:
            Misc.tback('Hub.run', e)


class NubDict(OrderedDict):
    """ Arrange for access to a dictionary to be annotated. """

    def __init__(self, name):
        OrderedDict.__init__(self)
        self.name = name
        self.listSelf()

    def __setitem__(self, k, v):
        OrderedDict.__setitem__(self, k, v)

        self.listSelf()

    def __delitem__(self, k):
        # k.shutdown(notifyHub=False)
        OrderedDict.__delitem__(self, k)

        self.listSelf()

    def listSelf(self, cmd=None):
        names = []
        for n in self.values():
            names.append(Misc.qstr(n.name))

        if not cmd:
            cmd = g.hubcmd

        cmd.inform('%s=%s' % (self.name, ','.join(names)))


class CmdrDict(NubDict):
    """ Like NubDict, but generate a 'users' keyword, depending on
    the state of the nub's isUser attribute.
    """

    def listSelf(self, cmd=None):
        if not cmd:
            cmd = g.hubcmd
        names = []
        userNames = []
        for n in self.values():
            names.append(Misc.qstr(n.name))
            if n.isUser:
                userNames.append(Misc.qstr(n.name))
            Misc.log('listCommanders', 'n=%s has info=%s' % (n, n.userInfo))
            if n.userInfo:
                cmd.inform(n.userInfo)

        cmd.inform('%s=%s' % (self.name, ','.join(names)))
        cmd.inform('users=%s' % (','.join(userNames)))


def addNubToDict(nub, nubDict):
    """Add a new nub to the given dict.
    The nub's ID must not be the same as for any existing nub. """

    existingNub = findNubInDict(nub.ID, nubDict)
    if existingNub is not None:
        Misc.log('Hub.nubs', 'nub %s already exists; not overwriting' % (nub.ID))
        return

    nubDict[nub.ID] = nub
    Misc.log('Hub.nubs', 'added nub %s to %s' % (nub, nubDict))


def dropNubFromDict(nub, nubDict, doShutdown=True):
    """ Close an existing nub. The nub must, of course, be in the given nub dict. """

    Misc.log('Hub.nubs', 'dropping nub=%s shutdown=%s' % (nub, doShutdown))
    if doShutdown:
        nub.shutdown(notifyHub=False)

    nub = findNubInDict(nub.ID, nubDict)
    if nub is None:
        Misc.log('Hub.nubs', 'nub %s is not registered; not dropping it' % (nub.ID))
        return

    del nubDict[nub.ID]


def findNubInDict(id, nubDict):
    """ Return the named nub, or None if it does not exist. """

    return nubDict.get(id, None)


def addActor(nub):
    addNubToDict(nub, g.actors)
    g.KVs.addSource(nub.name)
    if nub.needsAuth:
        g.perms.addActors([nub.name])


def dropActor(nub):
    # g.perms.dropActors([nub.name])
    g.KVs.clearSource(nub.name)
    dropNubFromDict(nub, g.actors)


def findActor(id):
    return findNubInDict(id, g.actors)


def addCommander(nub):
    Misc.log('hub.addCommander', 'adding %s' % (nub.name))
    addNubToDict(nub, g.commanders)


def dropCommander(nub, doShutdown=True):
    Misc.log('hub.dropCommander', 'dropping %s' % (nub.name))
    dropNubFromDict(nub, g.commanders, doShutdown=doShutdown)


def findCommander(id):
    return findNubInDict(id, g.commanders)


def addAcceptor(nub):
    addNubToDict(nub, g.acceptors)


def dropAcceptor(nub):
    dropNubFromDict(nub, g.acceptors)


def findAcceptor(id):
    return findNubInDict(id, g.acceptors)


def findNub(nub):
    """ Find whether a nub exists in any of the nub dictionaries. """

    for d in (g.actors, g.commanders, g.acceptors):
        if nub in d:
            return d[nub]
    return None


def dropNub(nub):
    """ Drop a Nub, regardless of its type. """

    if nub.ID in g.actors:
        dropActor(nub)
    elif nub.ID in g.commanders:
        dropCommander(nub)
    elif nub.ID in g.acceptors:
        dropAcceptor(nub)
    else:
        Misc.log('hub.dropNub',
                 'nub %s (%s) is neither in g.actors (%s) or g.commanders (%s)' %
                 (nub.ID, nub, g.actors, g.commanders))


def listActors(match):
    """ """

    actors = sorted(g.actors.keys())

    return actors


def validateCommanderNames(nub, programName, username):
    """ Transform a proposed CommanderNub name into a unique CommanderNub name.

    A Commander name must consist of a program name and a username separated by a period.
    Basically, take the proposed name and add _%d until there is no collision.
    """

    # Severely normalize the username
    #
    programName = re.sub('[^a-zA-Z0-9_]+', '_', programName)
    username = re.sub('[^a-zA-Z0-9_]+', '_', username)

    if re.match('[a-zA-Z_]', username) is None:
        username = '_' + username
    if re.match('[a-zA-Z_]', programName) is None:
        programName = '_' + programName

    fullName = '%s.%s' % (programName, username)

    n = 2
    proposedName = fullName
    ok = False
    while not ok:
        ok = True
        for c in g.commanders.values():
            if c.name == proposedName:
                ok = False
                break
        if not ok:
            proposedName = '%s_%d' % (fullName, n)
            n += 1

    return proposedName


def getActor(cmd):
    """ Find either an actor or vocabulary word matching the given command.
    """

    actorName = cmd.actorName

    # Misc.log("hub.getActor", "looking for actor %s" % (actorName))
    tgt = g.actors.get(actorName, None)

    if tgt is None:
        # Misc.log("hub.getActor", "looking for vocabulary word %s" % (actorName))
        tgt = g.vocabulary.get(actorName, None)

    Misc.log('hub.getActor', 'actornName %s target = %s' % (actorName, tgt))
    return tgt


def addCommand(cmd):
    """ Add a new command, and arrange for it to be executed by the appropriate target.

    Need to add:
        Automatically start an Actor if it is not running. Do this by:
          - check for starteable Nub - just call startNub in a catch.
          - send (i.e. queue) the command. We need to add the logic for the Actor's
        queue to be flushed when the connection has been established.

          - Provide some throttling to avoid idiocy when, say, the tcc password is changed.
    """

    Misc.log('hub.addCommand', 'new cmd=%s' % (cmd))

    if cmd.actorName == 'dbg':
        runCmd(cmd)
        return

    actor = getActor(cmd)

    if actor is None:
        cmd.fail('NoTarget=%s' %
                 Misc.qstr('the target named %s is not connected' % (cmd.actorName)),
                 src='hub')
        return

    # Checks whether the commander requires authentication
    if cmd.cmdr().needsAuth:
        # Enforce permissions if the actor requires them.
        ok = g.perms.checkAccess(cmd.cmdrCid, actor, cmd)
        if not ok:
            cmd.fail('NoPermission=%s' %
                     Misc.qstr('you do not have permission to command %s %s' %
                               (actor.name, cmd.cmd)),
                     src='hub')
            return

    actor.sendCommand(cmd)


def runCmd(c):
    cmd = c.cmd.strip()
    Misc.log('hub.runCmd', 'cmd = %r' % (cmd))
    if cmd == '':
        c.finish('Eval=%s' % (Misc.qstr('')),
                 src='hub')
        return

    try:
        ret = eval(cmd)
    except Exception as e:
        c.fail('EvalError=%s' % Misc.qstr(e),
               src='hub')
        raise

    c.finish('Eval=%s' % (Misc.qstr(ret)), src='hub')
    Misc.log('hub.runCmd', 'ret = %r' % (ret))


def forceReload(name):
    """ Do whatever we can to force a given module/package to be reloaded.

    """
    try:
        Misc.log('hub.forceReload', 'trying to (re-)load %s' % (name))
        mod = importlib.import_module(f'tron.Nubs.{name}')
    except BaseException:
        raise

    return mod


def stopNub(name):
    n = findActor(name)
    if n:
        dropActor(n)


def startNub(name):
    """Launch a single Nub.

    (Re-)Loads a module named 'name' from the Nubs folder and calls the start function.
    """

    Misc.log('hub.startNub', 'trying to start %s' % (name))

    mod = importlib.import_module(f'tron.Nubs.{name}')

    # And call the start() function.
    #
    Misc.log('hub.startNub', 'starting Nub %s...' % (name))
    mod.start(g.poller)


def loadPrograms(programsFile=None):
    """Loads the programs with their specific permissiosns."""

    if programsFile is None:
        programsFile = os.path.join(os.path.dirname(__file__), 'programs.json')

    if not os.path.exists(programsFile):
        Misc.log('hub.loadPrograms', 'programs file not found. Skipping.')
        return

    try:
        programs = json.load(open(programsFile))
    except BaseException:
        Misc.log('hub.loadPrograms', 'problem parsing JSON file.')
        return

    for program in programs:
        actors = programs[program]['actors']
        g.perms.addPrograms(programs=[program], actors=actors)



__all__ = ['RawReplyEncoder']

from tron import Misc
from tron.Hub.KV.KVDict import kvAsASCII
from tron.Parsing.dequote import dequote

from .ReplyEncoder import ReplyEncoder


class RawReplyEncoder(ReplyEncoder):
    """ Encode Replys without any protocol information.

    Options:
       EOL     - choose the EOL string to use.
       keyName - set a single keyword, whose _value_ alone will be returned.
    """

    def __init__(self, **argv):
        ReplyEncoder.__init__(self, **argv)

        # How do we terminate encoded lines?
        #
        self.EOL = argv.get('EOL', '\n')
        self.keyName = argv.get('keyName', None)

    def encode(self, r, nub, noKeys=False):
        """ Encode a protocol-free reply for a given nub.  """

        if self.keyName:
            rawVal = r.KVs.get(self.keyName, '')
            val = dequote(rawVal)
            Misc.log('RAWDEQUOTE', 'rawVal=%r val=%r' % (rawVal, val))
        else:
            val = self.encodeKeys(r.src, r.KVs)

        if val:
            return '%s%s' % (val, self.EOL)
        else:
            return ''

    def encodeKeys(self, src, KVs):
        """ Return a string encoding of KVs stored in an OrderedDict.

        Args:
           src   - ignored
           KVs   - an OrderedDict of values. See Parsing/parsing.py for important details.
        Notes:

        """

        if self.debug > 5:
            Misc.log('ASCIIReplyEnc.encode', 'encoding %r' % (KVs, ))
        if KVs is None:
            return ''

        keylist = []
        for k, v in KVs.items():
            if self.debug > 5:
                Misc.log('ASCIIReplyEnc.encode', 'encoding %r=%r' % (k, v))

            keylist.append(kvAsASCII(k, v))

        return '; '.join(keylist)



#!/usr/bin/env python

__all__ = ['PollHandler']

""" PollHandler.py -- wrap poll loop.

    The intent is to get non-blocking, event loop I/O.
    IOHandler instances can be registered to be checked for
    input or output. When IO is ready, the IOHandler callback is
    invoked. I.e. PollHandler does not read/write.
"""

import os
import select
import time
from threading import Lock

from tron import Misc


class NullIO(object):
    """ A Trick class to allow forcing the PollHandler to re-configure its outputs.
    """

    def __init__(self, fd, **argv):
        self.fd = fd
        self.debug = argv.get('debug', 0)

    def getInputFd(self):
        return self.fd

    def readInput(self):
        """ Pseudo-callback used to make the poller reconfigure itself. Does not need
        to actually do anything.

        """

        if self.debug > 0:
            Misc.log('NullIO', 'reading token')

        os.read(self.fd, 1)


class PollHandler(Misc.Object):
    """ Wrap the poll() system call.

        IOHandler instances get registered/unregistered for input
        and/or output. When input is available, or non-blocking output
        is possible, an IOHandler callback is triggered.

        Note that PollHandler does no I/O itself.

        Also, poll() acts on file descriptions, not files. Because it returns fds, which
        sort of determines how the callbacks are found, we only deal in fds internally.

        """

    def __init__(self, **argv):
        Misc.Object.__init__(self, **argv)

        self.poller = select.poll()

        self.files = {}
        self.lock = Lock()

        self.timedCallbacks = []
        self.cbLock = Lock()

        self.timeout = argv.get('timeout', 0.5)
        self.timeoutHandler = argv.get('timeoutHandler', None)

        # If there is any possibility that the polling list will be
        # changed during the poll() call proper, we need to wake the poller
        # up to re-read its list.
        #
        self.loopback = None
        self.looper = None
        needLoopback = argv.get('threaded', False)
        if needLoopback:
            self.startLoopback()

    def __del__(self):

        self.poller = None
        Misc.Object.__del__(self)

    def addTimer(self, timer):
        """ Add a timer.

        Args:
            timer   - a dictionary containing:
                         'callback'    - the function to call as callback(timer)
                         'time'        - a time.time() value to try to call by.
        Returns:
           - whether we have arranged for the loop to be restarted.
        """

        self.cbLock.acquire()
        self.timedCallbacks.append((timer['time'], timer))
        self.timedCallbacks.sort()

        # Kick the loop if necessary
        #
        if self.loopback and self.timedCallbacks[0][1] == timer:
            os.write(self.loopback, 'I')
        self.cbLock.release()

    def removeTimer(self, timer):
        """ Remove an existing timer.

        """
        self.cbLock.acquire()
        self.timedCallbacks.remove((timer['time'], timer))
        self.cbLock.release()

    def callMeIn(self, callback, delay):
        """ Arrange to call callback after delay seconds. """

        self.cbLock.acquire()
        self.timedCallbacks.append((time.time() + delay, callback))
        self.timedCallbacks.sort()
        self.cbLock.release()

    def startLoopback(self):
        """ Create a pipe that the poller listens to, that we can write to when the
        poller's file lists change. Also create the object that consumes the input.
        """

        rFd, wFd = os.pipe()

        self.looper = NullIO(rFd, debug=self.debug)
        self.loopback = wFd

        self.addInput(self.looper)

    def addInput(self, obj):
        """ Register an IOHandler instance for input and callback. Return existing handler or None.

        Args:
           obj  - an IOHandler instance, which must provide .getInputFd() and .readInput() methods.

        Returns:
           - the previously registered IOHandler object for the given input file, or None.
        """

        fd = obj.getInputFd()
        if fd is None or fd == -1:
            Misc.log('IOHandler.addInput', 'fd for obj=%s was %s!' % (obj, fd))
            return

        self.lock.acquire()

        pollInfo = self.files.get(fd, None)
        if self.debug > 2:
            Misc.log('Poll.registry',
                     'adding input for fd=%r obj=%s info=%r' % (fd, obj, pollInfo))

        if pollInfo:
            eventMask = pollInfo.get('eventMask', 0)
            lastHandler = pollInfo.get('putHandler', None)
            eventMask |= select.POLLIN
        else:
            lastHandler = None
            eventMask = select.POLLIN | select.POLLPRI
            pollInfo = {}
            self.files[fd] = pollInfo

        pollInfo['eventMask'] = eventMask
        pollInfo['inputHandler'] = obj
        self.poller.register(fd, eventMask)

        self.lock.release()

        # Wake the poller up.
        if self.loopback:
            os.write(self.loopback, 'I')

        if self.debug > 2:
            Misc.log(
                'Poll.registry',
                '%s added input %r(%s): %s' % (id(self), fd, self.flagNames(eventMask), repr(obj)))

        return lastHandler

    def addOutput(self, obj):
        """ Register an IOHandler instance for output and callback.

        Return existing handler or None.

        Args:
           obj  - an IOHandler instance, which must provide .getOutputFd()
                  and .canOutput() methods.

        Returns:
           - the previously registered IOHandler object for the given output file, or None.

        """
        fd = obj.getOutputFd()
        if fd is None or fd == -1:
            Misc.log('Poll.registry', 'Cannot add output for fd=%r' % (fd, ))
            return

        self.lock.acquire()
        pollInfo = self.files.get(fd, None)

        if self.debug > 2:
            Misc.log('Poll.registry',
                     'adding output for fd=%r obj=%s info=%r' % (fd, obj, pollInfo))

        if pollInfo:
            lastHandler = pollInfo.get('outputHandler', None)
            eventMask = pollInfo.get('eventMask', 0)
            eventMask |= select.POLLOUT

            changed = not eventMask & select.POLLOUT
        else:
            pollInfo = {}
            lastHandler = None
            eventMask = select.POLLOUT | select.POLLPRI
            self.files[fd] = pollInfo

            changed = True

        pollInfo['eventMask'] = eventMask
        pollInfo['outputHandler'] = obj
        self.poller.register(fd, eventMask)

        self.lock.release()

        # Wake the poller up.
        if self.loopback and changed:
            os.write(self.loopback, 'O')

        if self.debug > 2:
            Misc.log(
                'Poll.registry', '%s added output %r(%s): obj=%s info=%s' %
                (id(self), fd, self.flagNames(eventMask), obj, pollInfo))

        return lastHandler

    def removeInput(self, obj):
        return self.removeInputFd(obj.getInputFd())

    def removeInputFd(self, fd):
        """ Unregister an input. """

        if fd is None or fd == -1:
            Misc.log('Poll.registry', 'Cannot remove input for fd=%r' % (fd, ))
            return

        self.lock.acquire()

        pollInfo = self.files.get(fd, None)
        if self.debug > 2:
            Misc.log('Poll.registry', 'removing input for fd=%r info=%r' % (fd, pollInfo))

        if fd is None:
            Misc.log('Poll.registry', 'cannot change input for fd=None')
            self.lock.release()
            return

        if pollInfo:
            eventMask = pollInfo['eventMask']
            eventMask &= ~select.POLLIN
        else:
            Misc.log('Poll.registry',
                     'removeInput clearing all IO for unregistered object fd=%r.' % (fd))
            self.lock.release()
            return

        if pollInfo and eventMask != select.POLLPRI:
            pollInfo['eventMask'] = eventMask
            pollInfo['inputHandler'] = None
            self.poller.register(fd, eventMask)
            if self.debug > 2:
                Misc.log('Poll.registry',
                         'removed input %r and set mask to %s' % (fd, self.flagNames(eventMask)))
        else:
            Misc.log('Poll.registry', 'entirely removed (via in) fd=%s' % (fd))

            try:
                self.poller.unregister(fd)
            except Exception as e:
                Misc.log('Poll.registry',
                         'removeInput poller could not unregister fd=%s err=%s' % (fd, e))
            try:
                del self.files[fd]
            except Exception as e:
                Misc.log('Poll.registry', 'removeInput could not delete fd=%s err=%s' % (fd, e))

        self.lock.release()

        # Wake the poller up.
        if self.loopback:
            os.write(self.loopback, 'i')

    def removeOutput(self, obj):
        return self.removeOutputFd(obj.getOutputFd())

    def removeOutputFd(self, fd):
        """ Unregister an output. """

        self.lock.acquire()

        pollInfo = self.files.get(fd, None)
        if self.debug > 2:
            Misc.log('Poll.registry', 'removing output for fd=%r info=%r' % (fd, pollInfo))

        if fd is None:
            Misc.log('Poll.registry', 'cannot change output for fd=None')
            self.lock.release()
            return

        if pollInfo:
            eventMask = pollInfo['eventMask']
            eventMask &= ~select.POLLOUT
        else:
            Misc.log('Poll.registry',
                     'removeOutput clearing all IO for unregistered object fd=%r' % (fd))
            self.lock.release()
            return

        if pollInfo and eventMask != select.POLLPRI:
            pollInfo['eventMask'] = eventMask
            pollInfo['outputHandler'] = None
            self.poller.register(fd, eventMask)

            if self.debug > 2:
                Misc.log(
                    'Poll.registry',
                    'removed output fd=%s and set mask to %s' % (fd, self.flagNames(eventMask)))
        else:
            Misc.log('Poll.registry', 'entirely removing (via out) fd=%s' % (fd))
            try:
                self.poller.unregister(fd)
            except Exception as e:
                Misc.log('Poll.registry',
                         'removeOutput poller could not unregister fd=%s err=%s' % (fd, e))

            try:
                del self.files[fd]
            except Exception as e:
                Misc.log('Poll.registry', 'removeOutput could not delete fd=%s err=%s' % (fd, e))

        self.lock.release()

        # Wake the poller up.
        if self.loopback:
            os.write(self.loopback, 'o')

    def flagNames(self, flags):
        """ Return a string describing a poll event flag mask. """

        names = None
        for fmask, fname in ((select.POLLHUP, 'HUP'), (select.POLLERR, 'ERR'),
                             (select.POLLIN, 'IN'), (select.POLLOUT, 'OUT'),
                             (select.POLLNVAL, 'NVAL'), (select.POLLPRI, 'PRI')):
            if flags & fmask:
                if names is None:
                    names = fname
                else:
                    names += '|%s' % (fname, )

                flags &= ~fmask

        if flags != 0:
            if names is None:
                names = '%0x' % (flags, )
            else:
                names += '|%0x' % (flags, )

        return names

    def fileNames(self):
        """ Returns string describing the files we believe we are waiting on... """

        dlist = []
        for fd, f in self.files.items():
            inHandler = f.get('inputHandler', None)
            outHandler = f.get('outputHandler', None)

            dlist.append('fd=%s io=%s in=%s out=%s' %
                         (fd, self.flagNames(f['eventMask']), inHandler, outHandler))

        return ', '.join(dlist)

    def run(self):
        """ Wait for I/O and dispatch to handlers. """

        Misc.log('PollHandler.run', 'running...')

        while True:
            if self.debug > 7:
                Misc.log('PollHandler.run',
                         'loop, threaded=%s, id=%s' % (bool(self.looper is not None), id(self)))
                if self.debug > 8:
                    Misc.log('PollHandler.run', 'files=%s' % (self.fileNames()))

            # Calculate the proper timeout. Basically, use the loop default
            # or the next item in .timedCallbacks
            timeout = self.timeout
            self.cbLock.acquire()
            if self.timedCallbacks != []:
                nextTick, nextTimer = self.timedCallbacks[0]

                now = time.time()
                if nextTick - now < self.timeout:
                    timeout = nextTick - now
                    if timeout < 0.0:
                        timeout = 0.001
            self.cbLock.release()

            try:
                events = self.poller.poll(timeout * 1000.0)
            except (IOError, OSError) as e:
                Misc.log('PollHandler.run', 'poll trying to clean up: %s' % (e, ))
                try:
                    fd, eString = e
                    self.removeOutputFd(fd)
                    self.removeInputFd(fd)
                except BaseException:
                    Misc.log('PollHandler.run',
                             'poll failed with unknown error exception: %s' % (e, ))
            except Exception as e:
                Misc.log('PollHandler.run', 'poll failed with: %s (%s)' % (e, type(e)))
                if isinstance(e, type((), )) and len(e) == 2:
                    Misc.log('PollHandler.run', 'poll trying to clean up mess: %s' % (e, ))
                    fd, errString = e
                    self.removeOutputFd(fd)
                    self.removeInputFd(fd)
                else:
                    raise

            # The timer expired before any events became available.
            #
            if events == []:
                if self.timeoutHandler:
                    self.timeoutHandler()
                else:
                    if self.debug > 8:
                        Misc.log('PollHandler.run', 'time out on poll, with no timeoutHandler!')

            # Regardless of whether we got here by timeout or by event,
            # check the timed callbacks for expired events.
            #
            if self.timedCallbacks != []:
                now = time.time()
                timers = []
                self.cbLock.acquire()
                for i in range(len(self.timedCallbacks)):
                    try:
                        tick, timer = self.timedCallbacks[i]
                    except IndexError:
                        break

                    if tick > now:
                        break
                    timers.append(timer)
                    del self.timedCallbacks[i]
                self.cbLock.release()

                for timer in timers:
                    timer['callback'](timer)

            # Walk through all new events, and fire on all of them. Round-robinning provides
            # some simple protection against the worst starvation.
            #
            for event in events:
                fd, flag = event

                if self.debug > 4:
                    Misc.log('PollHandler.run', 'got fd=%s events=%s' % (fd, self.flagNames(flag)))

                if flag & ~(select.POLLIN | select.POLLOUT):
                    Misc.log(
                        'PollHandler.run',
                        'poll got exception flags: fd=%r, flag=%s' % (fd, self.flagNames(flag)))

                try:
                    d = self.files[fd]
                except KeyError:
                    Misc.log('PollHandler.run', 'invalid file on poll: %s' % (repr(fd)))

                    continue

                # Generate output first. Unlikely to matter.
                #
                if flag & select.POLLOUT:
                    callbackObj = d['outputHandler']
                    callbackObj.mayOutput()

                if flag & select.POLLIN:
                    callbackObj = d['inputHandler']
                    callbackObj.readInput()

                # Check exception flags separately from RW flags.
                # Why? Because there may have been I/O
                # pending before the error was raised.
                # Think of a client that closes right after writing.
                #
                # This is all a sad misunderstanding.
                # The original intent was to have this .run() loop
                # handle essentially all connection errors and closes.
                # But it turns out that Unixes vary
                # tremendously on how much poll() sees. In some cases,
                # HUP and ERR are never seen for
                # network sockets. Because of that, I am shifting the burden
                # to the callbacks -- read() and write()
                # do dependably generate errors.
                #
                if flag & (select.POLLHUP | select.POLLERR):
                    # On HUP or ERR, let the readInput() or mayOutput()
                    # discover the error and act on it.
                    #
                    Misc.log('PollHandler.run',
                             'HUP/ERR (%s) on poll: %s' % (self.flagNames(flag), repr(fd)))
                    outputHandler = d.get('outputHandler', None)
                    inputHandler = d.get('inputHandler', None)
                    if outputHandler:
                        outputHandler.shutdown()
                    if inputHandler:
                        inputHandler.shutdown()

                if flag & select.POLLNVAL:
                    # I don't know what I'm doing here. -- Misc
                    #
                    Misc.log('PollHandler.run',
                             'NVAL (%s) on poll: %s' % (self.flagNames(flag), repr(fd)))

                    outputHandler = d.get('outputHandler', None)
                    inputHandler = d.get('inputHandler', None)
                    if outputHandler:
                        outputHandler.shutdown()
                    if inputHandler:
                        inputHandler.shutdown()

                    # OK, the IOHandler shutdown has been called, but it is possible that
                    # it was not able to clear our polling data.
                    #
                    self.removeOutputFd(fd)
                    self.removeInputFd(fd)



__all__ = ['dequote']


def dequote(s):
    """ Convert s as a possibly quoted string to an unquoted string.

    If s is quoted, strip the quotes and unescape internal quotes..
    """

    # Misc.log("DEQUOTE", "IN q=%r= s=%r=" % (q, s))

    if s is None:
        return None

    assert isinstance(s, str)

    if len(s) >= 2:
        c0 = s[0]
        if c0 not in ('"', "'") or s[-1] != c0:
            return s

        # OK, we have a quoted string. Now build a new string and strip quotes.
        #
        if c0 == '"':
            findQuote = '\\"'
            replaceQuote = '"'
        else:
            findQuote = "\\'"
            replaceQuote = "'"

        sNoQuotes = s[1:-1].replace(findQuote, replaceQuote)

        # Then strip escapes.
        #
        sNoQuotes = sNoQuotes.replace('\\\\', '\\')

        # Misc.log("DEQUOTE", "OUT q=%r= s=%r=" % (q, sNoQuotes))

        return sNoQuotes

    return s


if __name__ == '__main__':
    pairs = (('', ''), ('""', ''), (' ', ' '), ('"', '"'), ("'", "'"), ('\\', '\\'),
             ('"\\\\"', '\\'), ('"\""', '"'), ("'\''", "'"),
             (r'"\a\b\c\d\e\f\g\""', r'\a\b\c\d\e\f\g"'), (r'"abc\"def\\\\ghi"', r'abc"def\\ghi'))

    print('testing dequote...')
    for test in pairs:
        s0, s1 = test
        sx = dequote(s0)
        if sx != s1:
            print('mismatch: s0=%r s1=%r got=%r' % (s0, s1, sx))
        else:
            print('ok: %r -> %r' % (s0, sx))



""" Auth.py - maintain authorization tables: who can command which actors?
"""

__all__ = ['Auth']

from tron import Misc, g


class Auth(Misc.Object):
    """
    Maintain an authorization table between commanders and actors.

    Basics:
       - A certain number of actors are registered with this package.
    """

    def __init__(self, defaultCmd, **argv):
        Misc.Object.__init__(self, **argv)

        self.defaultCmd = defaultCmd
        self.programs = {}
        self.actors = {}                # The actors subject to permissions.
        self.lockedActors = {}

        self.hackOn = False
        self.gods = ('APO', 'OBSERVER', 'LCO')

        for a in ['perms']:
            self.actors[a] = True

        # Gods are special: They:
        #  - can _always_ command "perms"
        #  - can command newly added actors.
        #
        self.addPrograms(self.gods, ['perms'])
        self.status()

    def checkAccess(self, cmdr, actor, cmd=None):
        """Return whether the given cmdr may command the given actor.

        Rules:
           - If the actor is locked, only allow APO users to command it.
           - If the program is unknown, deny access.
           - Otherwise check the program's access list.
        """

        program = cmdr.split('.', 1)[0].upper()

        if self.debug > 5:
            Misc.log('Auth.checkAccess', 'checking %s (%s) -> %s' % (cmdr, program, actor))

        if not cmd:
            cmd = self.defaultCmd

        needsAuth = actor.needsAuth

        if not needsAuth:
            return True

        actorName = actor.name

        # We can lock actors that we ordinarily don't know about, so check .lockedActors first
        if actorName in self.lockedActors:
            access = program in self.gods
            if access:
                return True
            else:
                cmd.warn('text=%s' % (Misc.qstr('%s is locked by APO' % (actorName))))
                return False

        # If we don't know about an actor, let the command go through
        if actorName not in self.actors:
            if self.hackOn and self.debug > 1:
                Misc.log('Auth.checkAccess', 'unregistered actor %s' % (actorName))
            return True

        # If the command is declared safe by the actor, let it go though.
        safeCmds = actor.safeCmds
        if self.debug > 1:
            Misc.log('auth.checkAccess', "checking '%s' against %s" % (cmd.cmd, safeCmds))
        if safeCmds is not None and cmd.cmd is not None:
            if safeCmds.search(cmd.cmd):
                return True

        # Let the hub command anything. This is for initialization commands, etc.
        if program == 'hub':
            return True

        # But if we don't know about a _program_, block the command.
        try:
            accessList = self.programs[program]
            if self.debug > 5:
                Misc.log('Auth.checkAccess', 'cmdr %s accessList = %s' % (cmdr, accessList))

            ok = actorName in accessList

            if self.hackOn:
                Misc.log('Auth.checkAccess', 'actor hacked on, in accessList = %s' % (ok))
                return True
            else:
                return ok
        except KeyError:

            # For SDSS, if permissions are disabled do not generate annoying warnings.
            if self.hackOn:
                return True

            # In this case, send a warning to our .defaultCmd as well as to the affected cmd
            cmd.warn(
                'permsTxt=%s' %
                (Misc.qstr(
                    'Authorization table has no entry for program: %s' %
                    (program))))
            if cmd != self.defaultCmd:
                self.defaultCmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Authorization table has no entry for program: %s' %
                        (program))))
            if self.hackOn:
                Misc.log('Auth.checkAccess', 'unknown program %s' % (program))
                return True
            else:
                return False

    def status(self, cmd=None):
        """ Generate all our keys.

        Args:
          cmd       - if set, the command to reply to. Otherwise use our .defaultCmd

        """

        self.genActorsKey(cmd=cmd)
        self.genProgramsKey(cmd=cmd)
        self.genLockedKey(cmd=cmd)
        self.genAuthKeys(cmd=cmd)

    def genActorsKey(self, cmd=None):
        """ Generate key describing the actors we control access to.

        Args:
           cmd       - if set, the command to reply to. Otherwise use our .defaultCmd

        Notes:
           Never finishes the cmd.
        """

        if not cmd:
            cmd = self.defaultCmd

        actors = list(self.actors.keys())
        actors.remove('perms')
        actors.sort()
        actors = [Misc.qstr(x) for x in actors]

        # HACK - older TUIs do not like empty variables ("name=")
        if actors:
            cmd.inform('actors=%s' % (','.join(actors)))
        else:
            cmd.inform('actors')

    def genProgramsKey(self, cmd=None):
        """ Generate key describing the programs we control access to.

        Args:
           cmd       - if set, the command to reply to. Otherwise use our .defaultCmd

        Notes:
           Never finishes the cmd.
        """

        if not cmd:
            cmd = self.defaultCmd

        programs = sorted(self.programs.keys())
        programs = [Misc.qstr(x) for x in programs]

        # HACK - older TUIs do not like empty variables ("name=")
        if programs:
            cmd.inform('programs=%s' % (','.join(programs)))
        else:
            cmd.inform('programs')

    def genLockedKey(self, cmd=None):
        """ Generate key describing the locked actors.

        Args:
           cmd    - if set, the command to reply to. Otherwise use our .defaultCmd

        Notes:
           Never finishes the cmd.
        """

        if not cmd:
            cmd = self.defaultCmd

        actors = sorted(self.lockedActors.keys())
        actors = [Misc.qstr(x) for x in actors]

        # HACK - older TUIs do not like empty variables ("name=")
        if actors:
            cmd.inform('lockedActors=%s' % (','.join(actors)))
        else:
            cmd.inform('lockedActors')

    def addActors(self, actors, cmd=None):
        """ Register a list of actors to be controlled.

        """

        if not cmd:
            cmd = self.defaultCmd

        if self.debug > 3:
            Misc.log('auth.addActors', 'adding actors %s' % (actors))

        for a in actors:
            if a in self.actors:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Actor %s is already registered for access control.' %
                        (a))))
            else:
                self.actors[a] = True

        # Prime superusers to be able to command newly connected actors
        #
        for god in self.gods:
            self.addActorsToProgram(god, actors)

        self.genActorsKey(cmd=cmd)

    def setLockedActors(self, actors, cmd=None):
        """ Block non-APO users form commanding a list of actors.

        Any actors not in .actors will be ignored with a warning.
        This may not be the right behavior. We may
        actually want to allow actors we do not control to be locked.

        Args:
           actors   - if None, lock all .actors. Else clear all existing locks
                      and add the given actors.
        """

        if not cmd:
            cmd = self.defaultCmd

        if actors is None:
            actors = self.actors

        if self.debug > 3:
            Misc.log('auth.lockActor', 'locking actors %s' % (actors))

        self.lockedActors = {}
        for a in actors:
            if a not in self.actors:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Actor %s is not subject to permissions and will not be locked' %
                        (a))))
            else:
                self.lockedActors[a] = True

        self.genLockedKey(cmd=cmd)

    def lockActors(self, actors, cmd=None):
        """ Block non-APO users form commanding a list of actors.

        Any actors not in .actors will be ignored with a warning.
        This may not be the right behavior. We may
        actually want to allow actors we do not control to be locked.
        """

        if not cmd:
            cmd = self.defaultCmd

        if not actors:
            actors = self.actors

        if self.debug > 3:
            Misc.log('auth.lockActor', 'locking actors %s' % (actors))

        for a in actors:
            if a in self.lockedActors:
                cmd.warn('permsTxt=%s' % (Misc.qstr('Actor %s is already locked' % (a))))
            elif a not in self.actors:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Actor %s is not subject to permissions and will not be locked' %
                        (a))))
            else:
                self.lockedActors[a] = True

        self.genLockedKey(cmd=cmd)

    def unlockActors(self, actors, cmd=None):
        """ unblock non-APO users from commanding a list of actors.
        """

        if not cmd:
            cmd = self.defaultCmd

        if not actors:
            actors = self.lockedActors

        if self.debug > 3:
            Misc.log('auth.unlockActor', 'unlocking actors %s' % (actors))

        for a in actors:
            try:
                del self.lockedActors[a]
            except KeyError:
                cmd.warn('permsTxt=%s' % (Misc.qstr('Actor %s was not locked' % (a))))

        self.genLockedKey(cmd=cmd)

    def genAuthKeys(self, programs=None, cmd=None):
        """ Generate keys describing some or all programs.

        Args:
           cmd       - if set, the command to reply to. Otherwise use our .defaultCmd
           programs  - if set, a list of programs to generate keys for.
                       Otherwise describe all programs.

        Notes:
           Never finishes the cmd.
        """

        if not cmd:
            cmd = self.defaultCmd
        if not programs:
            programs = list(self.programs.keys())

        programs.sort()
        Misc.log('auth.genAuthKeys', 'listing programs: %s' % (programs))

        for prog in programs:
            try:
                pAuth = list(self.programs[prog].keys())
            except KeyError:
                raise Exception('No authorization entry found for program %s' % (prog))

            pAuth.sort()
            actors = [Misc.qstr(x) for x in pAuth]
            cmd.inform('authList=%s,%s' % (Misc.qstr(prog), ','.join(actors)))

    def addPrograms(self, programs=[], actors=[], cmd=None):
        """ Add a list program to the control list.

        Args:
          program   - a list of program names. Use all connected programs if empty.
          actors    - an optional list of actors that the commander can command.
        """

        if not cmd:
            cmd = self.defaultCmd

        if not programs:
            programs = []
            for name, cmdr in g.commanders.items():
                if cmdr.needsAuth and name not in self.programs:
                    programs.append(name)

        if self.debug > 3:
            Misc.log('auth.addPrograms',
                     'adding programs %s with actors=%s' % (programs, actors))

        for prog in programs:
            if prog in self.programs:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Program %s already has an authorization entry, '
                        'which will not be modified.' % (prog))))
                continue
            prog = prog.upper()
            self.programs[prog] = {}
            self.setActorsForProgram(prog, actors, cmd=cmd)
        self.genProgramsKey(cmd=cmd)

    def dropPrograms(self, programs=[], cmd=None):
        """ Remove a list of programs from the control list.

        Args:
           programs  - a list of programs to register for authorization. Adds all connected
                       programs if empty.
           cmd       - The command to reply to, or .defaultCmd
        """

        if not programs:
            programs = list(self.programs.keys())

        if self.debug > 3:
            Misc.log('auth.dropProgram', 'dropping programs %s' % (programs))

        if not cmd:
            cmd = self.defaultCmd

        for program in programs:
            if program in self.gods:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Super-%s cannot be removed from the list of authorized programs' %
                        (program))))
                self.genAuthKeys([program], cmd)
                continue
            try:
                del self.programs[program]
            except BaseException:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Program %s did not have an authorization entry, so could not be deleted' %
                        (program))))

        self.genProgramsKey(cmd=cmd)

    def setActorsForProgram(self, program, actors, cmd=None):
        """ Define the list of actors that a program can command.
        """

        if self.debug > 3:
            Misc.log('auth.setActors', 'setting actors for commander %s: %s' % (program, actors))

        if not cmd:
            cmd = self.defaultCmd

        if program not in self.programs:
            cmd.fail(
                'permsTxt=%s' %
                (Misc.qstr(
                    'Program %s did not have an authorization entry, so could not be set' %
                    (program))))
            return

        d = {}
        for a in actors:
            d[a] = True
            if a not in self.actors:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Actor %s is not currentlu in the list of actors.' %
                        (a))))

        # Make sure God-like commanders can always command us...
        if program in self.gods:
            d['perms'] = True

        self.programs[program] = d

        self.genAuthKeys(programs=[program], cmd=cmd)

    def addActorsToProgram(self, program, actors, cmd=None):
        """ Add a list of actors to a actor's authorized list
        """

        if not cmd:
            cmd = self.defaultCmd

        if self.debug > 3:
            Misc.log('auth.addActors', 'adding actors for program %s: %s' % (program, actors))

        try:
            d = self.programs[program]
        except KeyError:
            cmd.fail(
                'permsTxt=%s' %
                (Misc.qstr(
                    'Program %s did not have an authorization entry, so could not be added to' %
                    (program))))
            return

        for a in actors:
            if a not in self.actors:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        'Actor %s is not subject to permissions.' %
                        (a))))
            else:
                d[a] = True

        self.genAuthKeys(programs=[program], cmd=cmd)

    def dropActorsFromProgram(self, program, actors, cmd=None):
        """ Remove a list of actors to a commander's authorized list.
        """

        if self.debug > 3:
            Misc.log('auth.dropActors', 'dropping actors for program %s: %s' % (program, actors))

        try:
            d = self.programs[program]
        except KeyError:
            cmd.fail(
                'permsTxt=%s' %
                (Misc.qstr(
                    'Program %s did not have an authorization entry, so could not be modified' %
                    (program))))
            return

        for a in actors:
            try:
                del d[a]
            except KeyError:
                cmd.warn(
                    'permsTxt=%s' %
                    (Misc.qstr(
                        "Actor %s was not in program %s's athorized list" %
                        (a, program))))

        self.genAuthKeys(programs=[program], cmd=cmd)


if __name__ == '__main__':
    def checkAndPrint(cmdr, actor, expect):
        access = a.checkAccess(cmdr, actor)
        if access == expect:
            ok = '    '
        else:
            ok = 'BAD '

        print('%s %s\t-> %s\t = %s' % (ok, cmdr, actor, access))

    g.actors = ('tcc', 'them')
    a = Auth(Misc.tcmd(name='auth'), debug=9)

    a.status()

    checkAndPrint('meme', 'them', False)

    a.addPrograms(['me'], ['yy'])
    checkAndPrint('meme', 'them', False)
    checkAndPrint('me.me', 'them', False)

    # a.addActors('xxxx', 'yyyy')
    a.addActors('me.me', ['them', 'theOthers'])
    checkAndPrint('me.me', 'them', True)
    checkAndPrint('me.me', 'theOthers', True)
    checkAndPrint('me.me', 'theOtherOtherss', False)

    a.addActors('me.me', ['them'])
    checkAndPrint('me.me', 'them', True)

    a.status()

    a.lockActors(['vvv', 'them'])
    checkAndPrint('me.me', 'them', False)

    a.addPrograms(['too'], ['them'])

    a.status()

    a.unlockActors(['them'])
    checkAndPrint('me.me', 'them', True)

    a.unlockActors(['them'])
    checkAndPrint('me.me', 'them', True)

    a.unlockActors(['vvv'])
    checkAndPrint('me.me', 'them', True)

    a.status()

    a.dropActors('me.me', ['them'])
    checkAndPrint('me.me', 'them', False)
    checkAndPrint('me.me', 'theOthers', True)

    a.dropPrograms([])
    checkAndPrint('me.me', 'theOthers', False)

    a.status()



""" A simple configuration manager. Loads python files in a given directory. That's it.
"""

__all__ = ['init', 'get', 'flush']

import json
import os
import sys

from tron.Misc.Exceptions import ICCError


cfgCache = None


def init(path=None, verbose=True):
    """ Initialize the cfg space.

    Args:
        path     - the directory inside which all the cfg files are kept.
    """

    global cfgPath

    if path is None:
        path = os.environ.get('CONFIG_DIR', None)

    if path is None:
        raise RuntimeError('Cannot initialize configuration module: no path given')

    cfgPath = path
    flush()

    if verbose:
        sys.stderr.write('initialized configuration under %s\n' % (cfgPath))


def flush():
    """ Clear any existing configuration cache.
    """

    global cfgCache

    cfgCache = {}


__nodef = 'no such variable HERE'


def get(space, var, default=__nodef, doFlush=False):
    """ Fetch a configuration value.

    Args:
        space     - the namespace to search.
        var       - the name of the variable to get.
        default   ? if set, and var is not in space, return this.
        doFlush   ? if True, reload the config cache.
    """

    if doFlush:
        flush()
    if cfgCache is None:
        init()

    try:
        s = cfgCache[space]
    except BaseException:
        s = _loadSpace(space)

    if id(default) == id(__nodef):
        return s[var]
    else:
        return s.get(var, default)


def _loadSpace(space):
    """ Load a configuration file into the cache.

    Args:
        space    - a namespace to load from cfgPath/space + ".py"
    """

    ldict = {}

    filename = os.path.join(cfgPath, '%s.json' % (space.lower()))
    try:
        ldict = json.load(open(filename))
    except Exception as e:
        raise ICCError('failed to read the configuration file %s: %s' % (filename, e))

    cfgCache[space] = ldict
    return ldict


def _test():
    init('/tmp/cfg')
    print(get('t1', 'x'))
    print(get('t1', 'x2'))


if __name__ == '__main__':
    _test()



__all__ = ['eatAVee', 'eatAString', 'parseKV', 'parseKVs', 'parseASCIIReply', 'parseRawReply']
""" Parsing utilities.

- Keywords can have zero or more comma-delimited values.
- The values are either quoted strings or unparsed tokens.
- Strings can be single- or double-quote delimited with internal quotes
  backslash-escaped. Non-escape backslashes within a string need to be
  backslash-escaped themselves.

- When parsed from a string, keywords are delimited by semicolons.

"""

import re
from collections import OrderedDict

from tron import Misc

from .Exceptions import ParseException


def eatAVee(s):
    """ Match a keyword value -- a possibly space-padded value ended by a
    whitespace, a comma, or a semicolon.

    Args:
       s - a string

    Returns:
      - the matched value. None if s is just whitespace.
      - any unmatched input, including the terminating character.
    """

    # Misc.log('eatAVee', 'called with %s' % (s))

    s = s.lstrip()
    if len(s) == 0:
        return '', ''

    # BEWARE - accept an empty value.
    if s[0] == ';':
        return '', s

    # String parsing is trickier, let eatAString() handle that.
    if s[0] in "\"'":
        return eatAString(s)

    vEnd = len(s)
    for i in range(len(s)):
        if s[i] in ';, \t\r\n\x0b\x0c':
            vEnd = i
            break

    if vEnd == 0:
        return '', s[vEnd:]

    return s[:vEnd], s[vEnd:]


def eatAString(s):
    """ Match a quote-escaped string.

    Args:
      s - a string, which must begin with a singor- or double- quote.

    Returns:
      - the matched string, or the rest of the line. The quotes are NOT removed.
      - any unmatched input, including the terminating character.

    NOTE:
      If the end of the input string is hit before the leading quote is closed, a closing quote
      is (fairly) silently appended.

    """

    # Misc.log('eatAString', 'called with %s' % (s))

    if len(s) == 0:
        raise ParseException('unexpected empty string while parsing', leftoverText='')

    startQuote = s[0]
    if startQuote != "\"" and startQuote != "\'":
        raise ParseException('string does not start with a quote', leftoverText=s)

    c = startQuote
    escaping = False
    for i in range(1, len(s)):
        if escaping:
            escaping = False
            continue

        c = s[i]

        if c == startQuote:
            return s[:i + 1], s[i + 1:]
        if c == '\\':
            escaping = True

    # OK, we fell off the end of the string without matching the closing quote.
    # Force the string to look OK so that nobody else needs to deal with a mangled string.
    #
    Misc.log('eatAString', 'adding closing section (esc=%s) to string %r' % (escaping, s))

    if escaping:
        s = '%s\\%s' % (s, startQuote)
    else:
        s += startQuote

    return s, ''


# Match "  key = STUFF"
kv_re = re.compile(
    r"""
  ^\s*                          # Ignore leading space
  (?P<key>[a-z_][a-z0-9_-]*)    # Match keyword name
  \s*                           # Ignore spaces after keyname
  (?P<delimiter>[=;]|$)
  \s*
  (?P<rest>.*)                  # Match eveything after the delimiter""",
    re.IGNORECASE | re.VERBOSE)


def parseKV(s):
    """ Try to parse a single KV.

    Return:
      { None, None, None } on end-of-input
      { K None rest-of-line } for a valueless keyword or
      { K V rest-of-line }
    """

    s = s.lstrip()
    if s == '':
        return None, None, None

    # Try to match for K=V. If we can't, try parsing as a valueless keyword.
    #
    match = kv_re.match(s)
    if match is None:
        raise ParseException(leftoverText=s)

    d = match.groupdict()
    K = d['key']
    rest = d['rest']

    # No equal sign? A valueless keyword.
    #
    if d['delimiter'] != '=':
        return K, None, rest

    # Build a list of values.
    #
    values = []
    while len(rest) != 0:

        # Parse a (sub-)value
        #
        try:
            V, rest = eatAVee(rest)
        except ParseException as e:
            e.prependText(rest)
            raise

        # BEWARE -- eatAVee() can return None for an empty value
        values.append(V)

        # Bail out if we:
        #   - hit EOL
        #
        rest = rest.lstrip()
        if len(rest) == 0:
            break

        # Keep gathering subvalues while we find commas.
        #
        if rest[0] == ',':
            rest = rest[1:]
        elif rest[0] == ';':
            rest = rest[1:]
            break
        else:
            break

    # Flatten singleton lists.
    if len(values) == 1:
        values = values[0]

    return K, values, rest


def parseKVs(s):
    """ Parse a string of key-value pairs into an OrderedDict .

    Returns:
      - an OrderedDict of keyword values.

    If a keyword has no value, the value is None
    Otherwise the value is a list of parsed values. Note that each value can be None.

    """

    KVs = OrderedDict()
    rest = s

    while True:
        try:
            key, values, rest = parseKV(rest)
        except ParseException as e:
            e.setKVs(KVs)
            raise

        if key is None:
            break

        # Misc.log('parseKVs', 'key=%r val=%r rest=%r' % (key, values, rest))
        KVs[key] = values

    return KVs


line_midcid_re = re.compile(
    r"""
  \s*                          # Skip leading whitespace
  (?P<mid>\d+)                 # integer MID
  \s+
  (?P<cid>[a-z0-9_][a-z0-9_.]*)                 # Integer CID. Should be more forgiving.
  \s+
  (?P<flag>[diwe:f>!])           # The flag. Should allow more characters,
                                 # and check them elsewhere.
  (?P<rest>.*)""", re.VERBOSE | re.IGNORECASE)

line_cidmid_re = re.compile(
    r"""
  \s*                          # Skip leading whitespace
  (?P<cid>[a-z0-9_][a-z0-9_.]*)  # Integer CID. Should be more forgiving.
  \s+
  (?P<mid>\d+)                 # integer MID
  \s+
  (?P<flag>[diwe:f>!])           # The flag. Should allow more characters,
                                 # and check them elsewhere.
  (?P<rest>.*)""", re.VERBOSE | re.IGNORECASE)


def parseASCIIReply(s, cidFirst=False):
    """ Try to parse a string into a dictionary containing:
         - mid   - the ICC's MID
         - cid   - the ICC's CID
         - flag  - the reply's flag character
         - KVs   - an OrderedDict of (key, value)s

        Returns that dictionary, or raises RuntimeError.

        If a reply line cannot be parsed at all, insert the entire line into the key 'RawLine'.
        If a reply line cannot be completely parsed, insert the unparsed section
        into the key 'UNPARSEDTEXT'.
    """

    if cidFirst:
        match = line_cidmid_re.match(s)
    else:
        match = line_midcid_re.match(s)

    if match is None:
        d = {}
        d['mid'] = 0
        d['cid'] = 0  # or 'hub' or '.hub'?
        d['flag'] = 'w'
        d['RawText'] = s

        kvs = OrderedDict()
        kvs['RawLine'] = [Misc.qstr(s)]
        d['KVs'] = kvs
        return d

    d = match.groupdict()

    try:
        KVs = parseKVs(d['rest'])
    except ParseException as e:
        KVs = e.KVs
        leftoverText = e.leftoverText

        # In this case, quote the offending text.
        KVs['UNPARSEDTEXT'] = [Misc.qstr(leftoverText)]
    except Exception as e:
        Misc.log('parseASCIIReply', 'unexpected Exception: %s' % (e))
        KVs = OrderedDict()
        KVs['UNPARSEDTEXT'] = [Misc.qstr(d['rest'])]

    d['KVs'] = KVs
    d['RawText'] = s
    del d['rest']

    return d


def parseRawReply(s, keyName='RawText'):
    """ Return a Reply with the entire input string saved in the given keyName keyword.

    Does not make any effort to determine command success, termination, failure, etc.
    """

    d = {}
    d['mid'] = 0
    d['cid'] = 0
    d['flag'] = 'i'

    Misc.log('parseRawReply', 'consumed :%r:' % (s))

    kvs = OrderedDict()
    kvs[keyName] = [Misc.qstr(s)]
    d['KVs'] = kvs
    d['RawText'] = s
    return d


def testParsing():
    OKtests = ('', 'o', 'expose', 'expose boo', 'expose boo=', "expose boo='bar'",
               'expose a b c=def h', 'e 1 2 3', "msg let me 'abc def' xx=123 yy='oh let me be'")

    NGtests = ("'", "abc'=1", "shortString='abcd", 'eol=2,3,')

    for t in OKtests:
        r = parseKVs(t)
        print('OKtest = %s' % (t))
        print('output = %s' % (r))
        print()

    for t in NGtests:
        print('NGtest = %s' % (t))
        try:
            r = parseKVs(t)
            print('output = %s' % (r))
        except Exception as e:
            print('exception = %s' % (e))

        print()


def testMatching():
    pass


if __name__ == '__main__':
    testParsing()
    testMatching()
